import multiprocessing
from pathlib import Path

from learna_tools.learna.agent import NetworkConfig, get_network, AgentConfig, ppo_agent_kwargs, get_agent
from learna_tools.learna.environment import RnaDesignEnvironment, RnaDesignEnvironmentConfig

from learna_tools.tensorforce.threaded_runner import clone_worker_agent, ThreadedRunner


def episode_finished(stats):
    """
    Function called after each episode of the agent (after designing an entire candidate solution).

    Args:
       stats: Statistics to be printed.

    Returns:
       True, meaning to continue running.
    """
    print(stats)
    return True


def learn_to_design_rna(
    dot_brackets,
    timeout,
    worker_count,
    save_path,
    restore_path,
    network_config,
    agent_config,
    env_config,
):
    """
    Main function for training the agent for RNA design. Instanciate agents and environments
    to run in a threaded runner using asynchronous parallel PPO.

    Args:
        TODO
        timeout: Maximum time to run.
        worker_count: The number of workers to run training on.
        save_path: Path for saving the trained model.
        restore_path: Path to restore saved configurations/models from.
        network_config: The configuration of the network.
        agent_config: The configuration of the agent.
        env_config: The configuration of the environment.

    Returns:
        Information on the episodes.
    """
    env_config.use_conv = any(map(lambda x: x > 1, network_config.conv_sizes))
    env_config.use_embedding = bool(network_config.embedding_size)
    environments = [
        RnaDesignEnvironment(dot_brackets, env_config) for _ in range(worker_count)
    ]

    network = get_network(network_config)
    agent = get_agent(
        environment=environments[0],
        network=network,
        agent_config=agent_config,
        session_config=None,
        restore_path=restore_path,
    )
    agents = clone_worker_agent(
        agent,
        worker_count,
        environments[0],
        network,
        ppo_agent_kwargs(agent_config, session_config=None),
    )
    threaded_runner = ThreadedRunner(agents, environments)
    # Bug in threaded runner requires a summary report
    threaded_runner.run(
        timeout=timeout, episode_finished=episode_finished, summary_report=lambda x: x
    )

    if save_path:
        save_path = Path(save_path)
        save_path.mkdir(parents=True, exist_ok=True)
        agent.save_model(directory=save_path.joinpath("last_model"))

    episodes_infos = [environment.episodes_info for environment in environments]
    return episodes_infos


if __name__ == "__main__":
    import argparse
    from ..data.parse_dot_brackets import parse_dot_brackets

    parser = argparse.ArgumentParser()

    # Data
    parser.add_argument("--data_dir", default="data", help="Data directory")
    parser.add_argument("--dataset", type=Path, help="Available: eterna, rfam_taneda")
    parser.add_argument(
        "--target_structure_ids",
        default=None,
        type=int,
        nargs="+",
        help="List of target structure ids to run on",
    )

    # Model
    parser.add_argument("--restore_path", type=Path, help="From where to load model")
    parser.add_argument("--save_path", type=Path, help="Where to save models")

    # Exectuion behaviour
    parser.add_argument("--timeout", type=int, help="Maximum time to run")
    parser.add_argument("--worker_count", type=int, help="Number of threads to use")

    # Hyperparameters
    parser.add_argument("--learning_rate", type=float, help="Learning rate to use")
    parser.add_argument(
        "--mutation_threshold", type=int, help="Enable MUTATION with set threshold"
    )
    parser.add_argument(
        "--reward_exponent", default=1, type=float, help="Exponent for reward shaping"
    )
    parser.add_argument(
        "--state_radius", default=0, type=int, help="Radius around current site"
    )
    parser.add_argument(
        "--conv_sizes", type=int, default=[1, 2], nargs="+", help="Size of conv kernels"
    )
    parser.add_argument(
        "--conv_channels",
        type=int,
        default=[50, 2],
        nargs="+",
        help="Channel size of conv",
    )
    parser.add_argument(
        "--num_fc_layers", type=int, default=2, help="Number of FC layers to use"
    )
    parser.add_argument(
        "--fc_units", type=int, default=50, help="Number of units to use per FC layer"
    )
    parser.add_argument(
        "--batch_size", type=int, default=8, help="Batch size for ppo agent"
    )
    parser.add_argument(
        "--entropy_regularization", type=float, default=1.5e-3, help="The output entropy"
    )
    parser.add_argument("--embedding_size", type=int, default=0, help="Size of embedding")
    parser.add_argument(
        "--lstm_units", type=int, default=0, help="Number of lstm units in each layer"
    )
    parser.add_argument(
        "--num_lstm_layers", type=int, default=0, help="Number of lstm layers"
    )

    args = parser.parse_args()

    network_config = NetworkConfig(
        conv_sizes=args.conv_sizes,
        conv_channels=args.conv_channels,
        num_fc_layers=args.num_fc_layers,
        fc_units=args.fc_units,
        embedding_size=args.embedding_size,
        lstm_units=args.lstm_units,
        num_lstm_layers=args.num_lstm_layers,
    )
    agent_config = AgentConfig(learning_rate=args.learning_rate)
    env_config = RnaDesignEnvironmentConfig(
        mutation_threshold=args.mutation_threshold,
        reward_exponent=args.reward_exponent,
        state_radius=args.state_radius,
    )
    dot_brackets = parse_dot_brackets(
        dataset=args.dataset,
        data_dir=args.data_dir,
        target_structure_ids=args.target_structure_ids,
    )
    learn_to_design_rna(
        dot_brackets,
        timeout=args.timeout,
        worker_count=args.worker_count or multiprocessing.cpu_count(),
        save_path=args.save_path,
        restore_path=args.restore_path,
        network_config=network_config,
        agent_config=agent_config,
        env_config=env_config,
    )



from pathlib import Path
import pandas as pd

def read_data(path):
    return path.glob('*.rna')

def generate_local_data(data):
    data = [(path.stem, path.read_text().rstrip()) for path in data]
    df = pd.DataFrame(data, columns=['id', 'structure'])
    df['sequence'] = df['structure']
    df['local_random'] = df['structure']
    df['local_motif'] = df['structure']
    df['gc_content'] = df['id']
    df['mfe'] = df['id']
    return df

def write_data(df, out_path):
    out_path.mkdir(parents=True, exist_ok=True)
    df.to_csv(path_or_buf=Path(out_path, 'test.interim'), index=False, sep='\t')

if __name__ == '__main__':
    path = Path('data', 'eterna')
    out_path = Path('data', 'eterna_local_test')

    data = read_data(path)

    df = generate_local_data(data)

    write_data(df, out_path)



import os

os.environ["OMP_NUM_THREADS"] = "1"
# os.environ['KMP_AFFINITY']='compact,1,0'

import multiprocessing

import numpy as np
import ConfigSpace as CS
from hpbandster.core.worker import Worker


from src.learna.agent import NetworkConfig, get_network, AgentConfig
from src.learna.environment import RnaDesignEnvironment, RnaDesignEnvironmentConfig
from src.learna.design_rna import design_rna


class DesignRNAWorker(Worker):
    def __init__(self, data_dir, num_cores, train_sequences, **kwargs):
        super().__init__(**kwargs)
        self.data_dir = data_dir
        self.train_sequences = train_sequences
        self.num_cores = num_cores

    def compute(self, config, budget, **kwargs):
        """
		Parameters
		----------
			budget: float
				cutoff for the agent on a single sequence
		"""

        config = self._fill_config(config)

        network_config = NetworkConfig(
            conv_sizes=[config["conv_size1"], config["conv_size2"]],
            conv_channels=[config["conv_channels1"], config["conv_channels2"]],
            num_fc_layers=config["num_fc_layers"],
            fc_units=config["fc_units"],
            num_lstm_layers=config["num_lstm_layers"],
            lstm_units=config["lstm_units"],
            fc_activation=config["fc_activation"],
            embedding_size=config["embedding_size"],
        )

        agent_config = AgentConfig(
            learning_rate=config["learning_rate"],
            batch_size=config["batch_size"],
            entropy_regularization=config["entropy_regularization"],
            likelihood_ratio_clipping=config["likelihood_ratio_clipping"],
        )

        environment_config = RnaDesignEnvironmentConfig(
            mutation_threshold=config["mutation_threshold"],
            include_mutation=config["mutation_threshold"] > 1,
            reward_exponent=config["reward_exponent"],
            state_radius=config["state_radius"],
        )

        validation_info = self._evaluate(
            "rfam_learn/validation",
            self.train_sequences,
            budget,
            config["restart_timeout"],
            network_config,
            agent_config,
            environment_config,
        )

        return {
            "loss": validation_info["sum_of_min_distances"],
            "info": {"validation_info": validation_info},
        }

    def _evaluate(
        self,
        dataset,
        evaluation_sequences,
        evaluation_timeout,
        restart_timeout,
        network_config,
        agent_config,
        environment_config,
    ):

        evaluation_arguments = [
            [
                dataset,
                self.data_dir,
                [i],  # target_structure_ids
                None,  # target_structure_path
                evaluation_timeout,  # timeout
                None,  # restore_path
                False,  # stop_learning
                restart_timeout,  # restart_timeout
                network_config,
                agent_config,
                environment_config,
            ]
            for i in evaluation_sequences
        ]

        with multiprocessing.Pool(self.num_cores) as pool:
            evaluation_results = pool.starmap(design_rna, evaluation_arguments)

        evaluation_sequence_infos = {}
        evaluation_sum_of_min_distances = 0
        evaluation_sum_of_first_distances = 0
        evaluation_num_solved = 0

        for r in evaluation_results:
            sequence_id = r[0].target_id
            r.sort(key=lambda e: e.time)

            times = np.array(list(map(lambda e: e.time, r)))
            dists = np.array(list(map(lambda e: e.fractional_hamming_distance, r)))

            evaluation_sum_of_min_distances += dists.min()
            evaluation_sum_of_first_distances += dists[0]

            evaluation_num_solved += dists.min() == 0.0

            evaluation_sequence_infos[sequence_id] = {
                "num_episodes": len(r),
                "mean_time_per_episode": float((times[1:] - times[:-1]).mean()),
                "min_distance": float(dists.min()),
                "last_distance": float(dists[-1]),
            }

        evaluation_info = {
            "num_solved": int(evaluation_num_solved),
            "sum_of_min_distances": float(evaluation_sum_of_min_distances),
            "sum_of_first_distances": float(evaluation_sum_of_first_distances),
            "squence_infos": evaluation_sequence_infos,
        }

        return evaluation_info

    @staticmethod
    def get_configspace():
        config_space = CS.ConfigurationSpace()

        # parameters for PPO here
        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "learning_rate", lower=1e-5, upper=1e-3, log=True, default_value=5e-4
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "batch_size", lower=32, upper=128, log=True, default_value=32
            )
        )
        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "entropy_regularization",
                lower=1e-5,
                upper=1e-2,
                log=True,
                default_value=1.5e-3,
            )
        )

        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "reward_exponent", lower=1, upper=10, default_value=1
            )
        )

        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "state_radius_relative", lower=0, upper=1, default_value=0
            )
        )

        # parameters for the architecture
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_radius1", lower=0, upper=8, default_value=1
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_channels1", lower=1, upper=32, log=True, default_value=32
            )
        )

        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_radius2", lower=0, upper=4, default_value=0
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_channels2", lower=1, upper=32, log=True, default_value=1
            )
        )

        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "num_fc_layers", lower=1, upper=2, default_value=2
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "fc_units", lower=8, upper=64, log=True, default_value=50
            )
        )

        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "num_lstm_layers", lower=0, upper=2, default_value=0
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "lstm_units", lower=1, upper=64, log=True, default_value=1
            )
        )

        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "embedding_size", lower=0, upper=4, default_value=1
            )
        )

        return config_space

    @staticmethod
    def _fill_config(config):
        config["mutation_threshold"] = 5

        config["conv_size1"] = 1 + 2 * config["conv_radius1"]
        if config["conv_radius1"] == 0:
            config["conv_size1"] = 0
        del config["conv_radius1"]

        config["conv_size2"] = 1 + 2 * config["conv_radius2"]
        if config["conv_radius2"] == 0:
            config["conv_size2"] = 0
        del config["conv_radius2"]

        min_state_radius = config["conv_size1"] + config["conv_size1"] - 1
        max_state_radius = 32
        config["state_radius"] = int(
            min_state_radius
            + (max_state_radius - min_state_radius) * config["state_radius_relative"]
        )

        config["likelihood_ratio_clipping"] = 0.3
        config["fc_activation"] = "relu"

        config["restart_timeout"] = None

        config["optimization_steps"] = 1

        return config



import os
import shutil
import multiprocessing


import numpy as np
import ConfigSpace as CS
from hpbandster.core.worker import Worker
from pathlib import Path

from src.learna.agent import NetworkConfig, get_network, AgentConfig
from src.learna.environment import RnaDesignEnvironment, RnaDesignEnvironmentConfig
from src.learna.design_rna import design_rna
from src.learna.learn_to_design_rna import learn_to_design_rna
from src.data.parse_dot_brackets import parse_dot_brackets, parse_local_design_data


class MetaLearnaAdaptWorker(Worker):
    def __init__(
        self, data_dir, num_cores, train_sequences, validation_timeout=60, **kwargs
    ):
        super().__init__(**kwargs)
        self.num_cores = num_cores
        self.validation_timeout = validation_timeout
        self._data_dir = data_dir
        self.sequence_ids = train_sequences

    def compute(self, config, budget, working_directory, config_id, **kwargs):
        """
		Parameters
		----------
			budget: float
				cutoff for the agent on a single sequence
		"""

        tmp_dir = os.path.join(
            working_directory, "%i_%i_%i" % (config_id[0], config_id[1], config_id[2])
        )
        shutil.rmtree(tmp_dir, ignore_errors=True)
        os.makedirs(tmp_dir, exist_ok=True)
        config = self._fill_config(config)

        if config["local_design"]:
            self.train_sequences = parse_local_design_data(
                dataset=Path(self._data_dir, config["trainingset"]).stem,
                data_dir=self._data_dir,
                target_structure_ids=self.sequence_ids,
            )
            self.validation_sequences = parse_local_design_data(
                dataset=Path(self._data_dir, "rfam_PD_validation").stem,
                data_dir=self._data_dir,
                target_structure_ids=range(1, 101),
            )
        else:
            self.train_sequences = parse_dot_brackets(
                dataset="rfam_learn/train",
                data_dir=self._data_dir,
                target_structure_ids=train_sequences,
            )
            self.validation_sequences = parse_dot_brackets(
                dataset="rfam_learn/validation",
                data_dir=self._data_dir,
                target_structure_ids=range(1, 101),
            )

        network_config = NetworkConfig(
            conv_sizes=[config["conv_size1"], config["conv_size2"]],
            conv_channels=[config["conv_channels1"], config["conv_channels2"]],
            num_fc_layers=config["num_fc_layers"],
            fc_units=config["fc_units"],
            num_lstm_layers=config["num_lstm_layers"],
            lstm_units=config["lstm_units"],
            embedding_size=config["embedding_size"],
            conv2d=config["encoding"] == 'multihot',
            embedding_indices=config["embedding_indices"]
        )

        agent_config = AgentConfig(
            learning_rate=config["learning_rate"],
            batch_size=config["batch_size"],
            entropy_regularization=config["entropy_regularization"],
        )

        env_config = RnaDesignEnvironmentConfig(
            reward_exponent=config["reward_exponent"],
            state_radius=config["state_radius"],
            local_design=config["local_design"],
            # sequence_reward=bool(config["sequence_reward"]),
            reward_function=config["reward_function"],
            predict_pairs=config["predict_pairs"],
            state_representation=config["state_representation"],
            data_type=config["data_type"],
            encoding=config["encoding"],
            # gc_weight=config["gc_weight"],
            # structural_weight=config["structural_weight"],
            # desired_gc=config["desired_gc"],
            # gc_improvement_step=True,
            # gc_postprocessing=False,
        )

        try:

            train_info = self._train(
                network_config, agent_config, env_config, tmp_dir, budget
            )
            validation_info = self._validate(
                network_config,
                agent_config,
                env_config,
                tmp_dir,
                config["restart_timeout"],
            )

        except:
            raise

        return {
            "loss": validation_info["sum_of_min_distances"],
            # "loss": validation_info["num_unsolved"],
            # "loss": validation_info["sum_of_min_deltas_and_distances"],
            "info": {"train_info": train_info, "validation_info": validation_info},
        }

    def _train(self, network_config, agent_config, env_config, tmp_dir, budget):

        # create arguments for all sequences
        train_arguments = [
            self.train_sequences,
            budget,  # timeout
            self.num_cores,  # worker_count
            tmp_dir,  # save_path
            None,  # restore_path
            network_config,
            agent_config,
            env_config,
        ]
        # need to run tensoflow in a separate thread otherwise the pool in _evaluate
        # does not work
        with multiprocessing.Pool(1) as pool:
            train_results = pool.apply(learn_to_design_rna, train_arguments)

        train_results = process_train_results(train_results)

        train_sequence_infos = {}
        train_sum_of_min_distances = 0
        train_sum_of_last_distances = 0
        train_num_solved = 0
        train_num_unsolved = 0

        for r in train_results.values():
            sequence_id = r[0].target_id
            r.sort(key=lambda e: e.time)

            dists = np.array(list(map(lambda e: e.normalized_hamming_distance, r)))

            train_sum_of_min_distances += dists.min()
            train_sum_of_last_distances += dists[-1]

            train_num_solved += dists.min() == 0.0
            train_num_unsolved += dists.min() != 0.0

            train_sequence_infos[sequence_id] = {
                "num_episodes": len(r),
                "min_distance": float(dists.min()),
                "last_distance": float(dists[-1]),
            }

        train_info = {
            "num_solved": int(train_num_solved),
            "num_unsolved": int(train_num_unsolved),
            "sum_of_min_distances": float(train_sum_of_min_distances),
            "sum_of_last_distances": float(train_sum_of_last_distances),
            "squence_infos": train_sequence_infos,
        }

        return train_info

    def _validate(
        self,
        network_config,
        agent_config,
        env_config,
        tmp_dir,
        restart_timeout,
        stop_learning=False,
    ):
        print("evaluating test performance")
        evaluation_arguments = [
            [
                [validation_sequence],
                self.validation_timeout,  # timeout
                tmp_dir,  # restore_path
                stop_learning,  # stop_learning
                restart_timeout,  # restart_timeout
                network_config,
                agent_config,
                env_config,
            ]
            for validation_sequence in self.validation_sequences
        ]

        with multiprocessing.Pool(self.num_cores) as pool:
            evaluation_results = pool.starmap(design_rna, evaluation_arguments)

        evaluation_sequence_infos = {}
        evaluation_sum_of_min_distances = 0
        evaluation_sum_of_first_distances = 0
        evaluation_num_solved = 0
        evaluation_num_unsolved = 0
        # evaluation_sum_of_min_gc_deltas = 0
        # evaluation_sum_of_min_gc_deltas_and_distances = 0


        for r in evaluation_results:
            sequence_id = r[0].target_id
            r.sort(key=lambda e: e.time)

            # times = np.array(list(map(lambda e: e.time, r)))
            dists = np.array(list(map(lambda e: e.normalized_hamming_distance, r)))
            # deltas_gc = np.array(list(map(lambda e: e.delta_gc, r)))
            # deltas_and_distances = np.array(list(map(lambda e: e.delta_gc + e.normalized_hamming_distance, r)))

            evaluation_sum_of_min_distances += dists.min()
            evaluation_sum_of_first_distances += dists[0]

            # evaluation_sum_of_min_gc_deltas += deltas_gc.min()
            # evaluation_sum_of_min_gc_deltas_and_distances += deltas_and_distances.min()

            evaluation_num_solved += dists.min() == 0.0
            evaluation_num_unsolved += dists.min() != 0.0

            evaluation_sequence_infos[sequence_id] = {
                "num_episodes": len(r),
                # "mean_time_per_episode": float((times[1:] - times[:-1]).mean()),
                "min_distance": float(dists.min()),
                "last_distance": float(dists[-1]),
                # "min_delta_gc": float(deltas_gc.min()),
                # "last_delta_gc": float(deltas_gc[-1]),
                # "min_deltas_and_distances": float(deltas_and_distances.min()),
                # "last_deltas_and_distances": float(deltas_and_distances[-1]),
            }

        evaluation_info = {
            "num_solved": int(evaluation_num_solved),
            "num_unsolved": int(evaluation_num_unsolved),
            "sum_of_min_distances": float(evaluation_sum_of_min_distances),
            "sum_of_first_distances": float(evaluation_sum_of_first_distances),
            # "sum_of_min_gc_deltas": float(evaluation_sum_of_min_gc_deltas),
            # "sum_of_min_deltas_and_distances": float(evaluation_sum_of_min_gc_deltas_and_distances),
            "squence_infos": evaluation_sequence_infos,
        }

        return evaluation_info

    @staticmethod
    def get_configspace():
        config_space = CS.ConfigurationSpace()

        # parameters for PPO here
        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "learning_rate", lower=1e-6, upper=1e-3, log=True, default_value=5e-4  # FR: changed learning rate lower from 1e-5 to 1e-6, ICLR: Learna (5,99e-4), Meta-LEARNA (6.44e-5)
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "batch_size", lower=32, upper=256, log=True, default_value=32  # FR: changed batch size upper from 128 to 256, configs from ICLR used 126 (LEARNA) and 123 (Meta-LEARNA)
            )
        )
        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "entropy_regularization",
                lower=1e-7,  # FR: changed entropy regularization lower from 1e-5 to 1e-7, ICLR: LEARNA (6,76e-5), Meta-LEARNA (151e-4)
                upper=1e-2,
                log=True,
                default_value=1.5e-3,
            )
        )

        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "reward_exponent", lower=1, upper=12, default_value=1  # FR: changed reward_exponent upper from 10 to 12, ICLR: Learna (9.34), Meta-LEARNA (8.93)
            )
        )

        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "state_radius_relative", lower=0, upper=1, default_value=0
            )
        )

        # parameters for the architecture
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_radius1", lower=0, upper=8, default_value=1
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_channels1", lower=1, upper=32, log=True, default_value=32
            )
        )

        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_radius2", lower=0, upper=4, default_value=0
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_channels2", lower=1, upper=32, log=True, default_value=1
            )
        )

        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "num_fc_layers", lower=1, upper=2, default_value=2
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "fc_units", lower=8, upper=64, log=True, default_value=50
            )
        )

        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "num_lstm_layers", lower=0, upper=3, default_value=0  # FR: changed lstm layers upper from 2 to 3
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "lstm_units", lower=1, upper=64, log=True, default_value=1
            )
        )

        # config_space.add_hyperparameter(
        #     CS.CategoricalHyperparameter(
        #         "reward_function", choices=['sequence_and_structure', 'structure_replace_sequence', 'structure_only']
        #     )
        # )

        config_space.add_hyperparameter(
            CS.CategoricalHyperparameter(
                "state_representation", choices=['n-gram', 'sequence_progress']
            )
        )

        config_space.add_hyperparameter(
            CS.CategoricalHyperparameter(
                "trainingset", choices=['rfam_LD_short_train', 'rfam_LD_train', 'rfam_LD_long_train', 'rfam_PD_short_train', 'rfam_PD_train', 'rfam_PD_long_train']
            )
        )

        config_space.add_hyperparameter(
            CS.CategoricalHyperparameter(
                "data_type", choices=['random', 'random-sort']
            )
        )


        encoding = CS.CategoricalHyperparameter(
                "encoding", choices=['sequence_bias', 'multihot', 'tuple']
            )
        embedding_size_seq_bias = CS.UniformIntegerHyperparameter(
                "embedding_size_seq_bias", lower=1, upper=9, default_value=1  # FR: changed embedding size upper from 4 to 8 and then to 9
            )
        embedding_size_tuple = CS.UniformIntegerHyperparameter(
                "embedding_size_tuple", lower=1, upper=21, default_value=1
            )
        config_space.add_hyperparameters([encoding, embedding_size_seq_bias, embedding_size_tuple])
        cond_tuple = CS.EqualsCondition(embedding_size_tuple, encoding, 'tuple')
        cond_seq_bias = CS.EqualsCondition(embedding_size_seq_bias, encoding, 'sequence_bias')
        config_space.add_condition(cond_tuple)
        config_space.add_condition(cond_seq_bias)


        return config_space

    @staticmethod
    def _fill_config(config):
        config["conv_size1"] = 1 + 2 * config["conv_radius1"]
        if config["conv_radius1"] == 0:
            config["conv_size1"] = 0
        del config["conv_radius1"]

        config["conv_size2"] = 1 + 2 * config["conv_radius2"]
        if config["conv_radius2"] == 0:
            config["conv_size2"] = 0
        del config["conv_radius2"]

        if config["conv_size1"] != 0:
            min_state_radius = config["conv_size1"] + config["conv_size1"] - 1
            max_state_radius = 32  # FR changed max state radius from 32 to 64, ICLR: LEARNA (32), Meta-LEARNA (29)
            config["state_radius"] = int(
                min_state_radius
                + (max_state_radius - min_state_radius) * config["state_radius_relative"]
            )
            del config["state_radius_relative"]
        else:
            min_state_radius = config["conv_size2"] + config["conv_size2"] - 1
            max_state_radius = 32  # FR changed max state radius from 32 to 64, ICLR: LEARNA (32), Meta-LEARNA (29)
            config["state_radius"] = int(
                min_state_radius
                + (max_state_radius - min_state_radius) * config["state_radius_relative"]
            )
            del config["state_radius_relative"]

        # config["desired_gc"] = np.random.choice([.1, .2, .3, .4, .5, .6, .7, .8, .9])
        config["restart_timeout"] = None
        config["local_design"] = True
        config["reward_function"] = 'structure_only'
        config["predict_pairs"] = 1
        if config["encoding"] == 'multihot':
            config["embedding_size"] = 0
            config["embedding_indices"] = 0
        elif config["encoding"] == 'tuple':
            config["embedding_size"] = config["embedding_size_tuple"]
            config["embedding_indices"] = 21
            del config["embedding_size_tuple"]
        else:
            config["embedding_size"] = config["embedding_size_seq_bias"]
            config["embedding_indices"] = 9
            del config["embedding_size_seq_bias"]


        # config["trainingset"] = 'rfam_PD_short_train'

        return config


def process_train_results(train_results):
    results_by_sequence = {}
    for r in train_results:
        for s in r:
            if not s.target_id in results_by_sequence:
                results_by_sequence[s.target_id] = [s]
            else:
                results_by_sequence[s.target_id].append(s)

    return results_by_sequence



import time
import math

from itertools import product
from dataclasses import dataclass
from distance import hamming
from collections import Counter

import numpy as np
from tensorforce.environments import Environment

from RNA import fold


def sequence_diversity_loss(count, n=3):
    """
    Maps the raw counts of appearances of the same sequences to the interval [0,1].

    Using the same formula as Alphafold2 for the cluster_deletion_value feature:

    2/\pi * arctan(d/3), where d are the raw counts.
    """
    # count = len(env_config['predictions'][prediction])

    return (2 / math.pi) * math.atan((count / n))



@dataclass
class RnaDesignEnvironmentConfig:
    """
    Dataclass for the configuration of the environment.

    Default values describe:
        mutation_threshold: Defines the minimum distance needed before applying the local
            improvement step.
        reward_exponent: A parameter to shape the reward function.
        state_radius: The state representation is a (2*<state_radius> + 1)-gram
            at each position.
        use_conv: Bool to state if a convolutional network is used or not.
        use_embedding: Bool to state if embedding is used or not.
    """

    mutation_threshold: int = 5
    reward_exponent: float = 1.0
    state_radius: int = 5
    use_conv: bool = True
    use_embedding: bool = False
    diversity_loss: bool = False


def _string_difference_indices(s1, s2):
    """
    Returns all indices where s1 and s2 differ.

    Args:
        s1: The first sequence.
        s2: The second sequence.

    Returns:
        List of indices where s1 and s2 differ.
    """
    return [index for index in range(len(s1)) if s1[index] != s2[index]]


def _encode_dot_bracket(secondary, env_config):
    """
    Encode the dot_bracket notated target structure. The encoding can either be binary
    or by the embedding layer.

    Args:
        secondary: The target structure in dot_bracket notation.
        env_config: The configuration of the environment.

    Returns:
        List of encoding for each site of the padded target structure.
    """
    padding = "=" * env_config.state_radius
    padded_secondary = padding + secondary + padding

    if env_config.use_embedding:
        site_encoding = {".": 0, "(": 1, ")": 2, "=": 3}
    else:
        site_encoding = {".": 0, "(": 1, ")": 1, "=": 0}

    # Sites corresponds to 1 pixel with 1 channel if convs are applied directly
    if env_config.use_conv and not env_config.use_embedding:
        return [[site_encoding[site]] for site in padded_secondary]
    return [site_encoding[site] for site in padded_secondary]


def _encode_pairing(secondary):
    """TODO
    """
    pairing_encoding = [None] * len(secondary)
    stack = []
    for index, symbol in enumerate(secondary, 0):
        if symbol == "(":
            stack.append(index)
        elif symbol == ")":
            paired_site = stack.pop()
            pairing_encoding[paired_site] = index
            pairing_encoding[index] = paired_site
    return pairing_encoding


class _Target(object):
    """TODO
    Class of the target structure. Provides encodings and id.
    """

    _id_counter = 0

    def __init__(self, dot_bracket, env_config, target_id=None):
        """
        Initialize a target structure.

        Args:
             dot_bracket: dot_bracket encoded target structure.
             env_config: The environment configuration.
        """
        _Target._id_counter += 1
        if target_id is None:
            self.id = _Target._id_counter  # For processing results
        else:
            self.id = target_id
        self.dot_bracket = dot_bracket
        self._pairing_encoding = _encode_pairing(self.dot_bracket)
        self.padded_encoding = _encode_dot_bracket(self.dot_bracket, env_config)

    def __len__(self):
        return len(self.dot_bracket)

    def get_paired_site(self, site):
        """
        Get the paired site for <site> (base pair).

        Args:
            site: The site to check the pairing site for.

        Returns:
            The site that pairs with <site> if exists.TODO
        """
        return self._pairing_encoding[site]


class _Design(object):
    """
    Class of the designed candidate solution.
    """

    action_to_base = {0: "G", 1: "A", 2: "U", 3: "C"}
    action_to_pair = {0: "GC", 1: "CG", 2: "AU", 3: "UA"}

    def __init__(self, length=None, primary=None):
        """
        Initialize a candidate solution.

        Args:
            length: The length of the candidate solution.
            primary: The sequence of the candidate solution.
        """
        if primary:
            self._primary_list = primary
        else:
            self._primary_list = [None] * length
        self._dot_bracket = None
        self._current_site = 0

    def get_mutated(self, mutations, sites):
        """
        Locally change the candidate solution.

        Args:
            mutations: Possible mutations for the specified sites
            sites: The sites to be mutated

        Returns:
            A Design object with the mutated candidate solution.
        """
        mutatedprimary = self._primary_list.copy()
        for site, mutation in zip(sites, mutations):
            mutatedprimary[site] = mutation
        return _Design(primary=mutatedprimary)

    def assign_sites(self, action, site, paired_site=None):
        """
        Assign nucleotides to sites for designing a candidate solution.

        Args:
            action: The agents action to assign a nucleotide.
            site: The site to which the nucleotide is assigned to.
            paired_site: defines if the site is assigned with a base pair or not.
        """
        self._current_site += 1
        if paired_site:
            base_current, base_paired = self.action_to_pair[action]
            self._primary_list[site] = base_current
            self._primary_list[paired_site] = base_paired
        else:
            self._primary_list[site] = self.action_to_base[action]

    @property
    def first_unassigned_site(self):
        try:
            while self._primary_list[self._current_site] is not None:
                self._current_site += 1
            return self._current_site
        except IndexError:
            return None

    @property
    def primary(self):
        return "".join(self._primary_list)


def _random_epoch_gen(data):
    """
    Generator to get epoch data.

    Args:
        data: The targets of the epoch
    """
    while True:
        for i in np.random.permutation(len(data)):
            yield data[i]


@dataclass
class EpisodeInfo:
    """
    Information class.
    """

    __slots__ = ["target_id",
                 "time",
                 "normalized_hamming_distance",
                 "hamming_distance",
                 "structure",
                 "sequence",
                 ]
    target_id: int
    time: float
    normalized_hamming_distance: float
    hamming_distance: int
    structure: str
    sequence: str


class RnaDesignEnvironment(Environment):
    """
    The environment for RNA design using deep reinforcement learning.
    """

    def __init__(self, dot_brackets, env_config):
        """TODO
        Initialize an environemnt.

        Args:
            env_config: The configuration of the environment.
        """
        self._env_config = env_config

        if isinstance(dot_brackets[0], str):
            targets = [_Target(dot_bracket, self._env_config) for dot_bracket in dot_brackets]
        else:
            targets = [_Target(dot_bracket, self._env_config, target_id=i) for i, dot_bracket in dot_brackets]
        self._target_gen = _random_epoch_gen(targets)

        self.target = None
        self.design = None
        self.episodes_info = []
        self.predictions = []

    def __str__(self):
        return "RnaDesignEnvironment"

    def seed(self, seed):
        return None

    def reset(self):
        """
        Reset the environment. First function called by runner. Returns first state.

        Returns:
            The first state.
        """
        self.target = next(self._target_gen)
        self.design = _Design(len(self.target))
        return self._get_state()

    def _apply_action(self, action):
        """
        Assign a nucleotide to a site.

        Args:
            action: The action chosen by the agent.
        """
        current_site = self.design.first_unassigned_site
        paired_site = self.target.get_paired_site(current_site)  # None for unpaired sites
        self.design.assign_sites(action, current_site, paired_site)

    def _get_state(self):
        """
        Get a state dependend on the padded encoding of the target structure.

        Returns:
            The next state.
        """
        start = self.design.first_unassigned_site
        return self.target.padded_encoding[
            start : start + 2 * self._env_config.state_radius + 1
        ]

    def _local_improvement(self, folded_design):
        """
        Compute Hamming distance of locally improved candidate solutions.

        Returns:
            The minimum Hamming distance of all imporved candidate solutions.
        """
        differing_sites = _string_difference_indices(
            self.target.dot_bracket, folded_design
        )
        candidates = []
        for mutation in product("AGCU", repeat=len(differing_sites)):
            mutated = self.design.get_mutated(mutation, differing_sites)
            folded_mutated, _ = fold(mutated.primary)
            hamming_distance = hamming(folded_mutated, self.target.dot_bracket)
            if hamming_distance == 0:  # For better timing results
                return 0, mutated, folded_mutated
            candidates.append((hamming_distance, mutated, folded_mutated))

        return min(candidates, key=lambda x: x[0])

    def _get_reward(self, terminal):
        """
        Compute the reward after assignment of all nucleotides.

        Args:
            terminal: Bool defining if final timestep is reached yet.

        Returns:
            The reward at the terminal timestep or 0 if not at the terminal timestep.
        """
        if not terminal:
            return 0
        primary = self.design.primary
        folded_design, _ = fold(primary)
        hamming_distance = hamming(folded_design, self.target.dot_bracket)
        if 0 < hamming_distance < self._env_config.mutation_threshold:
            hamming_distance, primary, folded_design  = self._local_improvement(folded_design)

        normalized_hamming_distance = hamming_distance / len(self.target)

        if self._env_config.diversity_loss:

            count = Counter(self.predictions)[primary]
            self.predictions.append(primary)
            div_loss = sequence_diversity_loss(count / len(self.predictions))

            loss = normalized_hamming_distance + div_loss

            normalized_distance = min(1.0, loss)



        normalized_distance = normalized_hamming_distance if not self._env_config.diversity_loss else normalized_distance
        # For hparam optimization
        episode_info = EpisodeInfo(
            target_id=self.target.id,
            time=time.time(),
            normalized_hamming_distance=normalized_distance,
            hamming_distance=hamming_distance,
            structure=folded_design,
            sequence=primary,
        )
        self.episodes_info.append(episode_info)

        return (1 - normalized_distance) ** self._env_config.reward_exponent

    def execute(self, actions):
        """
        Execute one interaction of the environment with the agent.

        Args:
            action: Current action of the agent.

        Returns:
            state: The next state for the agent.
            terminal: The signal for end of an episode.
            reward: The reward if at terminal timestep, else 0.
        """
        self._apply_action(actions)

        terminal = self.design.first_unassigned_site is None
        state = None if terminal else self._get_state()
        reward = self._get_reward(terminal)

        return state, terminal, reward

    def close(self):
        pass

    @property
    def states(self):
        type = "int" if self._env_config.use_embedding else "float"
        if self._env_config.use_conv and not self._env_config.use_embedding:
            return dict(type=type, shape=(1 + 2 * self._env_config.state_radius, 1))
        return dict(type=type, shape=(1 + 2 * self._env_config.state_radius,))

    @property
    def actions(self):
        return dict(type="int", num_actions=4)



import pandas as pd
import numpy as np
import string
from pathlib import Path
from ..learna.environment import _encode_pairing

class DatasetError(Exception):
    pass

def tsv_to_df(path):
    try:
        return pd.read_csv(Path(path), sep='\t')
    except:
        print(f"error with file {path}")

def drop_bad_sequences(df):
    valid_rows = [row for row in df.itertuples() if set(row.sequence).issubset('ACGU') and set(row.structure).issubset('.()')]
    print(f"{len(valid_rows)} valid targets of {df['sequence'].size} -- drop {df['sequence'].size - len(valid_rows)}")
    data = pd.DataFrame(valid_rows)
    return data

def validate_datasets(datasets):
    validation = []
    for dataset in datasets:
        validation.append(all([set(sequence).issubset('AGCU') for sequence in dataset['sequence']]))
    return all(validation)

def validate_structure_length(df):
    return all([x == len(y) for x, y in zip(df['length'], df['structure'])])

def split_structure_and_mfe(df):
    struc_and_mfe = df.structure_and_mfe.values.tolist()
    struc_and_mfe = [(str(x.strip("[]").split(', ')[0].strip("''")), float(x.strip("[]").split(', ')[1])) for x in struc_and_mfe]
    df[['structure', 'mfe']] = pd.DataFrame(struc_and_mfe, index=df.index)
    return df[['family', 'length', 'sequence', 'structure', 'mfe', 'gc_content']]

def remove_duplicates(df):
    return df.drop_duplicates(subset=['structure'], keep=False)

def satisfy_length_criteria(df, min, max):
    return df.loc[(df['length'] >= min) & (df['length'] <= max)]

def generate_datasets(df, size, train_multiplier, validation_multiplier):
    samples = size + (size * train_multiplier) + (size * validation_multiplier)
    if samples > df.size:
        raise DatasetError(f"Not enough entries in the database to create disjoint datasets\n database entries: {df.size}\n desired dataset size: {samples}")

    data = df.sample(n=samples)
    test_set = data.iloc[:size]
    training_set = data.iloc[size: size + (size * train_multiplier)]
    validation_set = data.iloc[size + (size * train_multiplier):]

    return [test_set, training_set, validation_set]

def generate_local_random(df, max_replacements, replace_quantile, suffix_size):
    local_structures = []

    for structure, sequence in zip(df['structure'], df['sequence']):
        suf_size = np.random.randint(1, suffix_size)
        replacements = np.random.randint(1, max_replacements)

        struc = [s for s in structure[:-suf_size]]

        current_site = 0
        for i in range(replacements):
            try:
                current_replacement_site = np.random.randint(1, int(replace_quantile * len(structure)))

                struc[current_site:current_site + current_replacement_site] = sequence[current_site:current_site + current_replacement_site]
                struc = ''.join(struc)
                struc = [c for c in struc]

                step = np.random.randint(1, int(replace_quantile * len(structure)))
                current_site += current_replacement_site + step

            except Exception as e:
                print('**********************************************************************')
                print(e)
                raise DatasetError(f"Error with structure {structure} and sequence {sequence}")
        struc = struc + sequence[len(struc):].split()
        struc = ''.join(struc)
        local_structures.append(struc)

    df['local_random'] = local_structures
    return df

def extract_motifs(df):
    print('Extract motifs from data')
    motifs = []
    for structure, sequence in zip(df['structure'], df['sequence']):
        pairing_encoding = _encode_pairing(structure)
        structure_motifs = [(structure[index:pair+1], sequence[index:pair+1], index, pair) for index, pair in enumerate(pairing_encoding) if pair and structure[index:pair+1]]
        motifs.append(structure_motifs)
    df['motifs'] = motifs
    return df


def generate_local_motif_based(df):
    print('Generate motif based random local structures')
    motif_based_local_structures = []
    for sequence, motifs in zip(df['sequence'], df['motifs']):
        try:
            motif = motifs[np.random.randint(0, len(motifs) - 2)]
            seq = sequence[:motif[2]] + motif[0] + sequence[motif[3] + 1:]
            motif_based_local_structures.append(seq)
        except:
            try:
                if len(motifs) > 2:
                    motif = motifs[np.random.randint(0, len(motifs) - 2)]
                    seq = sequence[:motif[2]] + motif[0] + sequence[motif[3] + 1:]
                    motif_based_local_structures.append(seq)
                else:
                    motif_based_local_structures.append(sequence)
                    print("Not enough motifs")
                    print("Use original sequence instead")

            except Exception as e:
                print('********************************************************************************')
                print(f"Please try again")

                raise DatasetError('Please restart.')

    df['local_motif'] = motif_based_local_structures
    return df

def datasets_to_tsv(datasets, out_dir, name, args):
    test_path = Path(out_dir, f"{name}_test")
    train_path = Path(out_dir, f"{name}_train")
    validation_path = Path(out_dir, f"{name}_validation")

    test_path.mkdir(parents=True, exist_ok=True)
    train_path.mkdir(parents=True, exist_ok=True)
    validation_path.mkdir(parents=True, exist_ok=True)

    print('Write test set data')
    datasets[0].to_csv(path_or_buf=Path(out_dir, f"{name}_test", "test.interim"), sep='\t', index=False)
    datasets[0]['structure'].to_csv(Path(out_dir, f"{name}_test", "test.rna"), sep='\t', index=False)
    datasets[0]['gc_content'].to_csv(Path(out_dir, f"{name}_test", "test.gc"), sep='\t', index=False)
    datasets[0]['mfe'].to_csv(Path(out_dir, f"{name}_test", "test.mfe"), sep='\t', index=False)
    datasets[0]['sequence'].to_csv(Path(out_dir, f"{name}_test", "test.seq"), sep='\t', index=False)

    if args.local_random:
        datasets[0]['local_random'].to_csv(Path(out_dir, f"{name}_test", "test.local"), sep='\t', index=False)
        datasets[0]['local_motif'].to_csv(Path(out_dir, f"{name}_test", "test.motif"), sep='\t', index=False)
    if args.motifs:
        datasets[0]['motifs'].to_csv(Path(out_dir, f"{name}_test", "test.motifs"), sep='\t', index=False)

    print('Write training set data')
    datasets[1].to_csv(path_or_buf=Path(out_dir, f"{name}_train", "train.interim"), sep='\t', index=False)
    datasets[1]['structure'].to_csv(Path(out_dir, f"{name}_train", "train.rna"), sep='\t', index=False)
    datasets[1]['mfe'].to_csv(Path(out_dir, f"{name}_train", "train.mfe"), sep='\t', index=False)
    datasets[1]['sequence'].to_csv(Path(out_dir, f"{name}_train", "train.seq"), sep='\t', index=False)
    datasets[1]['gc_content'].to_csv(Path(out_dir, f"{name}_train", "train.gc"), sep='\t', index=False)
    if args.local_random:
        datasets[1]['local_random'].to_csv(Path(out_dir, f"{name}_train", "train.local"), sep='\t', index=False)
        datasets[1]['local_motif'].to_csv(Path(out_dir, f"{name}_train", "train.motif"), sep='\t', index=False)
    if args.motifs:
        datasets[1]['motifs'].to_csv(Path(out_dir, f"{name}_train", "train.motifs"), sep='\t', index=False)

    print('Write validation set data')
    datasets[2].to_csv(path_or_buf=Path(out_dir, f"{name}_validation", "validation.interim"), sep='\t', index=False)
    datasets[2]['structure'].to_csv(Path(out_dir, f"{name}_validation", "validation.rna"), sep='\t', index=False)
    datasets[2]['mfe'].to_csv(Path(out_dir, f"{name}_validation", "validation.mfe"), sep='\t', index=False)
    datasets[2]['sequence'].to_csv(Path(out_dir, f"{name}_validation", "validation.seq"), sep='\t', index=False)
    datasets[2]['gc_content'].to_csv(Path(out_dir, f"{name}_validation", "validation.gc"), sep='\t', index=False)
    if args.local_random:
        datasets[2]['local_random'].to_csv(Path(out_dir, f"{name}_validation", "validation.local"), sep='\t', index=False)
        datasets[2]['local_motif'].to_csv(Path(out_dir, f"{name}_validation", "validation.motif"), sep='\t', index=False)
    if args.motifs:
        datasets[2]['motifs'].to_csv(Path(out_dir, f"{name}_validation", "validation.motifs"), sep='\t', index=False)


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()

    # Data
    parser.add_argument(
        "--tsv_path",
        default="data/rfam_interim",
        help="Path to interim tsv",
    )
    parser.add_argument("--out_dir", default="data/rfam_processed", help="Output path")

    # Dataset specification
    parser.add_argument(
        "--family",
        default=None,
        required=False,
        type=str,
        nargs="+",
        help="List of Rfam family ids to use for dataset. Format: <int> id of family, e.g., 1 3 180",
    )

    parser.add_argument(
        "--name", default="rfam_learn_local", required=False, type=str, help="Dataset name"
    )

    # Hyperparameters
    parser.add_argument("--size", type=int, default=100, help="The size of the test set")
    parser.add_argument(
        "--train_multiplier",
        required=False,
        type=int,
        default=1000,
        help="Trainig set size <multiplier> times <size> ",
    )
    parser.add_argument(
        "--validation_multiplier",
        required=False,
        type=int,
        default=1,
        help="Validation set size <multiplier> times <size> ",
    )

    parser.add_argument(
        "--minimum_length",
        required=False,
        type=int,
        default=None,
        help="The minimum sequence length of the targets"
    )

    parser.add_argument(
        "--maximum_length",
        required=False,
        type=int,
        default=None,
        help="The maximum sequence length of the targets"
    )

    parser.add_argument(
        "--unique",
        action="store_true",
        help="Only use unique structures for dataset",
    )

    parser.add_argument(
        "--local_random",
        action="store_true",
        help="Create dataset for local RNA Design",
    )

    parser.add_argument(
        "--motifs",
        action="store_true",
        help="Create motif based sequence-structure targets",
    )


    parser.add_argument(
        "--maximum_replacements",
        required=False,
        type=int,
        default=5,
        help="The number of replacements for random local sequences"
    )

    parser.add_argument(
        "--suffix_size",
        required=False,
        type=int,
        default=20,
        help="The number of replacements for random local sequences"
    )


    parser.add_argument(
        "--replace_quantile",
        required=False,
        type=float,
        default=0.3,
        help="The percentage of the structure that can be replaced with sequence for each replacement"
    )

    args = parser.parse_args()

    print(f"Read data from {args.tsv_path}")

    dfs = []

    # if certain families were chosen, look only at these, otherwise use all families
    if args.family:
        print("Satisfy family criteria")
        for family in args.family:
            id = ("0" * (5 - len(family))) + family
            dfs.append(tsv_to_df(Path(args.tsv_path, 'targets_' + id + '.tsv')))
    else:
        print('No family criteria provided -- Read all data')
        for path in Path(args.tsv_path).glob('*.tsv'):
            dfs.append(tsv_to_df(path))

    df = pd.concat(dfs)
    print(f"whole data : {df['sequence'].size}")

    # remove structures that aren't in the chosen length interval
    if args.maximum_length or args.minimum_length:
        print("Satisfy length criteria")
        min = 0 if args.minimum_length is None else args.minimum_length
        max = 100000000 if args.maximum_length is None else args.maximum_length
        df = satisfy_length_criteria(df, min, max)

    # split structure and mfe data into two columns (provided as list from RNAfold)
    df = split_structure_and_mfe(df)

    # only use unique sequences if desired by user
    if args.unique:
        print("Use unique structures only")
        df = remove_duplicates(df)

    print(f"Unique structures: {df['sequence'].size}")

    # print('Validate data')
    # assert validate_structure_length(df)
    # df = drop_bad_sequences(df)

    # print('Generate datasets')
    # print(f"test set size: {args.size}")
    # print(f"training set size: {args.size * args.train_multiplier}")
    # print(f"validation set size: {args.size * args.validation_multiplier}")
    # datasets = generate_datasets(df, args.size, args.train_multiplier, args.validation_multiplier)
    # iteration = 0
    # while not validate_datasets(datasets):
    #     iteration += 1
    #     print(f"datasets not valid in iteration {iteration}")
    #     datasets = generate_datasets(df, args.size, args.train_multiplier, args.validation_multiplier)

    # # generate randomly generated local design structures if desired
    # if args.local_random:
    #     print("Generate random structures for local design")
    #     datasets = [generate_local_random(dataset, args.maximum_replacements, args.replace_quantile, args.suffix_size) for dataset in datasets]

    # if args.motifs:
    #     # extract motifs from structures and generate local data by inserting structural motifs into sequence
    #     print('Generate motif based structures for local design ')
    #     datasets = [generate_local_motif_based(extract_motifs(dataset)) for dataset in datasets]

    # print(f"Write data to {args.out_dir}")
    # name = args.name if args.name else 'rfam_learn_local'
    # datasets_to_tsv(datasets, args.out_dir, name=name, args=args)



from pathlib import Path


def parse_dot_brackets(
    dataset, data_dir, target_structure_ids=None, target_structure_path=None
):
    """TODO
    Generate the targets for next epoch.

    Args:
        dataset: The name of the benchmark to use targets from.
        data_dir: The directory of the target structures.
        target_structure_ids: Use specific targets by ids.
        target path: Specify a path to the targets.

    Returns:
        An epoch generator for the specified target structures.
    """
    if target_structure_path:
        target_paths = [target_structure_path]
    elif target_structure_ids:
        target_paths = [
            Path(data_dir, dataset, f"{id_}.rna") for id_ in target_structure_ids
        ]
    else:
        target_paths = list(Path(data_dir, dataset).glob("*.rna"))

    return [data_path.read_text().rstrip() for data_path in target_paths]



import time
import re
import subprocess

import numpy as np
import pandas as pd

from itertools import product
from dataclasses import dataclass
from distance import hamming
from io import StringIO
from pathlib import Path
from Bio import AlignIO

from tensorforce.environments import Environment
from RNA import fold

# from src.e2efold.e2efold_productive_short import e2efold_main_short
# from src.e2efold.e2efold_productive_long import e2efold_main_long
# from src.e2efold.utils import ct2db


@dataclass
class RnaDesignEnvironmentConfig:
    """
    Dataclass for the configuration of the environment.

    Default values describe:
        mutation_threshold: Defines the minimum distance needed before applying the local
            improvement step.
        reward_exponent: A parameter to shape the reward function.
        state_radius: The state representation is a (2*<state_radius> + 1)-gram
            at each position.
        use_conv: Bool to state if a convolutional network is used or not.
        use_embedding: Bool to state if embedding is used or not.
        gc_improvement_step: Bool to decide if gc_improvement step is used.
        gc_tolerance: The tolerance for gc content.
        desired_gc: The gc_content that is desired.
        gc_weight: Determines how much weight is set on the gc content control.
        gc_reward: Bool to decide if the gc_content is included in the reward function.
        num_actions: The number of actions the agent can choose from.
    """
    mutation_threshold: int = 5
    reward_exponent: float = 1.0
    state_radius: int = 5
    use_conv: bool = False
    use_embedding: bool = False
    local_design: bool = False
    predict_pairs: bool = False
    reward_function: str = 'structure'
    state_representation: str = 'sequence-progress'
    data_type: str = 'random'
    sequence_constraints: str = None
    encoding: str = 'sequence_bias'
    min_length: int = 0
    max_length: int = 0
    min_dot_extension: int = 0
    variable_length: bool = False
    control_gc: bool = False
    desired_gc: float = None
    gc_tolerance: float = 0.01
    algorithm: str = 'rnafold'
    folding_counter: int = 0
    interactions: int = 0
    seed: int = 42  # 1  # 123 # 0  # 137  # 42
    cm_design: bool = False
    cm_name: str = 'RF00008'
    cm_path: str = 'rfam_cms/Rfam.cm'
    working_dir: str = 'working_dir'
    rri_design: bool = False
    rri_target: str = 'UUUAAAUUAAAAAAUCAUAGAAAAAGUAUCGUUUGAUACUUGUGAUUAUACUCAGUUAUACAGUAUCUUAAGGUGUUAUUAAUAGUGGUGAGGAGAAUUUAUGAAGCUUUUCAAAAGCUUGCUUGUGGCACCUGCAACUCUUGGUCUUUUAGCACCAAUGACCGCUACUGCUAAU'  # mRNA1 of intarna example



tuple_encoding = {
    ('=', '='): 0,
    ('A', '.'): 1,
    ('A', '('): 2,
    ('A', ')'): 3,
    ('A', 'N'): 4,
    ('C', '.'): 5,
    ('C', '('): 6,
    ('C', ')'): 7,
    ('C', 'N'): 8,
    ('G', '.'): 9,
    ('G', '('): 10,
    ('G', ')'): 11,
    ('G', 'N'): 12,
    ('U', '.'): 13,
    ('U', '('): 14,
    ('U', ')'): 15,
    ('U', 'N'): 16,
    ('N', '.'): 17,
    ('N', '('): 18,
    ('N', ')'): 19,
    ('N', 'N'): 20,
}

# tuple_encoding = {
#     ('=', '='): 0,
#     ('A', '.'): 1,
#     ('A', '('): 2,
#     ('A', ')'): 3,
#     # ('A', '|'): 4,
#     ('A', 'N'): 5,
#     ('C', '.'): 6,
#     ('C', '('): 7,
#     ('C', ')'): 8,
#     # ('C', '|'): 8,
#     ('C', 'N'): 9,
#     ('G', '.'): 10,
#     ('G', '('): 11,
#     ('G', ')'): 12,
#     # ('G', '|'): 13,
#     ('G', 'N'): 14,
#     ('U', '.'): 15,
#     ('U', '('): 16,
#     ('U', ')'): 17,
#     # ('U', '|'): 18,
#     ('U', 'N'): 19,
#     ('N', '.'): 20,
#     ('N', '('): 21,
#     ('N', ')'): 22,
#     # ('N', '|'): 23,
#     ('N', 'N'): 24,
# }



def _fold_primary(primary, env_config):
    # increment folding counter
    env_config.folding_counter += 1
    # fold with rnafold
    if env_config.algorithm == 'rnafold':
        folding = fold(primary)[0]
    elif env_config.algorithm == 'rnafold_mea':
        pathname = ''.join([str(env_config.desired_gc)])+'_rnafold.fasta'
        rnafold_path = Path(pathname)
        with open(rnafold_path, 'w+') as f:
            f.write('>current' + '\n')
            f.write(primary)
        folding = subprocess.Popen(["RNAfold", "--MEA", "-i",  str(rnafold_path.resolve())], stdout=subprocess.PIPE).communicate()[0]
        folding = folding.decode("utf-8").split('\n')[5].split()[0]  # 5 is MEA structure output, 0 is structure in DB notation
        rnafold_path.unlink()
    # fold with e2efold
    elif env_config.algorithm == 'e2efold':
        if len(primary) > 600:
            ct_df = pd.read_csv(StringIO(e2efold_main_long(primary, 'src/e2efold/config_long.json')), sep='\s+', header=None)
            folding = ''.join(change_encoding(ct2db(ct_df)))
        else:
            ct_df = pd.read_csv(StringIO(e2efold_main_short(primary, 'src/e2efold/config.json')), sep='\s+', header=None)
            folding = ''.join(change_encoding(ct2db(ct_df)))
    # fold with linearfold
    elif env_config.algorithm == 'linearfold':
        echo_ps = subprocess.Popen(['echo', f"{primary}"], stdout=subprocess.PIPE)
        folding = subprocess.Popen(["./thirdparty/linearfold/LinearFold/linearfold"], stdin=echo_ps.stdout, stdout=subprocess.PIPE).communicate()[0].decode("utf-8").split()[1]
    elif env_config.algorithm == 'linearfoldV':
        echo_ps = subprocess.Popen(['echo', f"{primary}"], stdout=subprocess.PIPE)
        folding = subprocess.Popen(["./thirdparty/linearfold/LinearFold/linearfold", "-V"], stdin=echo_ps.stdout, stdout=subprocess.PIPE).communicate()[0].decode("utf-8").split()[1]

    elif env_config.algorithm == 'contrafold':
        pathname = ''.join([str(env_config.desired_gc)])+'_contrafold.fasta'
        contrafold_path = Path(pathname)
        with open(contrafold_path, 'w+') as f:
            f.write('>current' + '\n')
            f.write(primary)
            # echo_ps = subprocess.Popen(['echo', ">current" + '\n' + f"{primary}"], stdout=f)
        folding = subprocess.Popen(["contrafold", "predict", "--noncomplementary", str(contrafold_path.resolve())], stdin=contrafold_path.open(), stdout=subprocess.PIPE).communicate()[0].decode("utf-8").split()[-1]  # , stdin=echo_ps.stdout
        contrafold_path.unlink()
    return folding


def _encode_tuple(secondary, primary, env_config):
    """
    Encodes the sequence and structure constraints into a joint represenation from tuples of sequence and structure.

    Args:
        secondary <str>: input structure constraints.
        primary <str>: input sequence constraints.
        env_config <RnaDesignEnvironmentConfig>: the environment config file.

    Returns:
        A list/list of lists that encodes each site with a natural number.
    """
    padding = "=" * env_config.state_radius
    padded_secondary = padding + secondary + padding
    padded_primary = padding + primary + padding

    p_primary_list = list(padded_primary)
    p_secondary_list = list(padded_secondary)

    p_encoding = [tuple_encoding[(x, y)] for x, y in zip(p_primary_list, p_secondary_list)]

    if env_config.use_conv and not env_config.use_embedding:
        return [[site] for site in p_encoding]
    return [site for site in p_encoding]


def multihot(secondary, primary, env_config):
    """TODO
    """
    # one_hot = {
    #     'A': np.array([[1], [0], [0], [0], [0], [0], [0], [0], [0], [0]]),  # [A, C, G, U, N, ., (, ), |, =]
    #     'C': np.array([[0], [1], [0], [0], [0], [0], [0], [0], [0], [0]]),
    #     'G': np.array([[0], [0], [1], [0], [0], [0], [0], [0], [0], [0]]),
    #     'U': np.array([[0], [0], [0], [1], [0], [0], [0], [0], [0], [0]]),
    #     'N': np.array([[0], [0], [0], [0], [1], [0], [0], [0], [0], [0]]),
    #     '.': np.array([[0], [0], [0], [0], [0], [1], [0], [0], [0], [0]]),
    #     '(': np.array([[0], [0], [0], [0], [0], [0], [1], [0], [0], [0]]),
    #     ')': np.array([[0], [0], [0], [0], [0], [0], [0], [1], [0], [0]]),
    #     '|': np.array([[0], [0], [0], [0], [0], [0], [0], [0], [1], [0]]),
    #     '=': np.array([[0], [0], [0], [0], [0], [0], [0], [0], [0], [1]]),
    # }

    one_hot = {
        'A': np.array([[1], [0], [0], [0], [0], [0], [0], [0], [0]]),  # [A, C, G, U, N, ., (, ), |, =]
        'C': np.array([[0], [1], [0], [0], [0], [0], [0], [0], [0]]),
        'G': np.array([[0], [0], [1], [0], [0], [0], [0], [0], [0]]),
        'U': np.array([[0], [0], [0], [1], [0], [0], [0], [0], [0]]),
        'N': np.array([[0], [0], [0], [0], [1], [0], [0], [0], [0]]),
        '.': np.array([[0], [0], [0], [0], [0], [1], [0], [0], [0]]),
        '(': np.array([[0], [0], [0], [0], [0], [0], [1], [0], [0]]),
        ')': np.array([[0], [0], [0], [0], [0], [0], [0], [1], [0]]),
        '=': np.array([[0], [0], [0], [0], [0], [0], [0], [0], [1]]),
    }


    padding = "=" * env_config.state_radius
    padded_secondary = padding + secondary + padding
    padded_primary = padding + primary + padding

    p_primary_list = list(padded_primary)
    p_secondary_list = list(padded_secondary)

    p_primary_encoding = list(map(lambda x: one_hot[x], p_primary_list))
    p_secondary_encoding = list(map(lambda x: one_hot[x], p_secondary_list))

    p_primary_encoding = np.stack(p_primary_encoding, axis=0)
    p_secondary_encoding = np.stack(p_secondary_encoding, axis=0)

    encoding = p_primary_encoding + p_secondary_encoding
    encoding[encoding > 1] = 1

    return encoding


def _string_difference_indices(s1, s2):
    """
    Returns all indices where s1 and s2 differ.

    Args:
        s1: The first sequence.
        s2: The second sequence.

    Returns:
        List of indices where s1 and s2 differ.
    """
    return [index for index in range(len(s1)) if s1[index] != s2[index]]


def _encode_dot_bracket(secondary, env_config):
    """
    Encode the dot_bracket notated target structure. The encoding can either be binary
    or by the embedding layer.

    Args:
        secondary: The target structure in dot_bracket notation.
        env_config: The configuration of the environment.

    Returns:
        List of encoding for each site of the padded target structure.
    """
    padding = "=" * env_config.state_radius
    padded_secondary = padding + secondary + padding

    if env_config.use_embedding:
        # site_encoding = {".": 0, "(": 1, ")": 2, "|": 3, "=": 4, "A": 5, "G": 6, "C": 7, "U": 8, "N": 9}
        site_encoding = {".": 0, "(": 1, ")": 2, "=": 3, "A": 4, "G": 5, "C": 6, "U": 7, "N": 8}
    else:
        site_encoding = {".": 0, "(": 0, ")": 0, "|": 0, "=": 1, "A": 2, "G": 2, "C": 2, "U": 2, "N": 1}

    # Sites corresponds to 1 pixel with 1 channel if convs are applied directly
    if env_config.use_conv and not env_config.use_embedding:
        return [[site_encoding[site]] for site in padded_secondary]
    return [site_encoding[site] for site in padded_secondary]


def _encode_pairing(secondary):
    """TODO
    """
    # initialize pairing encoding with None everywhere
    pairing_encoding = [None] * len(secondary)
    # initialize stack
    stack = []
    # loop over input secondary
    for index, symbol in enumerate(secondary, 0):
        # if opening bracket pu on stack; note opening brackets correspond to next closing bracket, therefore opening brackets on the stack don't matter for unbalanced brackets
        if symbol == "(":
            stack.append(index)
        # if closing bracket, pair with last opening bracket on stack; if there is no stack, closing bracket is unbalanced bracket and we continue
        elif symbol == ")":
            if not stack:
                continue
            # if there is a stack pair with last opening bracket
            else:
                paired_site = stack.pop()
                pairing_encoding[paired_site] = index
                pairing_encoding[index] = paired_site
    return pairing_encoding


def _encode_structure_parts(local_target):
    """TODO
    """
    encoding = [None] * len(local_target)
    for index, site in enumerate(local_target):
        if site in ['.', '(', ')', 'N']:
            encoding[index] = site
    return encoding


class GcControler(object):
    """TODO
    Class for local improvement of the GC content of the designed sequence.
    """
    def __init__(self, design, target, gc_tolerance, env_config):
        self._env_config = env_config
        self._design = design
        self._target = target
        self._gc_tolerance = gc_tolerance
        if self._target.gc is None:
            self._desired_gc = 0.5
        else:
            self._desired_gc = float(self._target.gc)

    def gc_content_primary(self, primary):
        return (primary.upper().count('G') + primary.upper().count('C')) / len(primary)

    def gc_content(self, design):
        return (design.primary.upper().count('G') + design.primary.upper().count('C')) / len(design.primary)

    def gc_diff(self, design):
        return self.gc_content(design) - self.desired_gc

    def gc_diff_abs(self, design):
        return abs(self.gc_diff(design))

    def gc_diff_abs_primary(self, primary):
        return abs(self.gc_content_primary(primary) - self.desired_gc)

    def gc_satisfied(self, design):
        return self.gc_diff_abs(design) <= self._gc_tolerance

    @property
    def gc_tolerance(self):
        return self._gc_tolerance

    @property
    def desired_gc(self):
        return self._desired_gc

    def gc_improvement_step(self, hamming_distance=0, site=0):
        """TODO
        """
        # if gc satisfied just return the design
        if self.gc_satisfied(self._design):
            return self._design

        # start with paired sites, save single sites for second gc improvement step if necessary
        single_sites = []

        sites = list(range(0, len(self._design.primary)))

        # go through all sites of the current best candidate solution (updated during improvement)
        for i in range(len(self._design.primary)):

            site = np.random.choice(sites)
            sites.remove(site)

            # if gc already satisfied return current design
            if self.gc_satisfied(self._design):
                return self._design

            # if site is constrained in sequence space, don't change it
            if self._target.sequence_constraints[site] != 'N':
                continue

            # if site is single site, ignore now, use later
            if not self._target.get_paired_site(site):
                single_sites.append(site)
                continue

            # if gc content is too low (gc_current - gc_desired) < 0, increas gc content
            if self.gc_diff(self._design) < 0:
                self._increase_gc(site, self._target.get_paired_site(site), hamming_distance)

            # else decreas gc content
            else:
                self._decrease_gc(site, self._target.get_paired_site(site), hamming_distance)

        # if improving paired sites wasn't sufficient, start improving single sites
        if not self.gc_satisfied(self._design):
            self._gc_improve_single_sites(single_sites, hamming_distance=hamming_distance)

        # return the current (improved design)
        return self._design


    def _gc_improve_single_sites(self, single_sites, hamming_distance=0):
        """TODO
        """
        sites = single_sites
        # walk over all single sites
        for i in range(len(single_sites)):

            site = np.random.choice(sites)
            sites.remove(site)

            # if gc content satisfied, break
            if self.gc_satisfied(self._design):
                break
            # should already not happen because checked in outer loop already...but...
            if self._target.sequence_constraints[site] != 'N':
                continue
            # else, start improving
            else:
                # if gc content is too low (gc_current - gc_desired) < 0, increas gc content
                if self.gc_diff(self._design) < 0:
                    self._increase_gc(site, self._target.get_paired_site(site), hamming_distance)
                # else decreas gc content
                else:
                    self._decrease_gc(site, self._target.get_paired_site(site), hamming_distance)

    def _increase_gc(self, site, paired_site=None, hamming_distance=0):
        """TODO
        """
        # create two temporary primaries to check for gc-improvement (two because we swap A/U with G and C)
        primary_tmp_1 = list(self._design.primary)
        primary_tmp_2 = list(self._design.primary)
        # if current site is A or U replace with G and C (and watson-crick pair base if site is paired)
        if self._design.primary[site] in ['A', 'U']:
            # assign G and C separately
            primary_tmp_1[site] = 'G'
            primary_tmp_2[site] = 'C'
            # and assign paired site if nucleotide is paired; paired site is not restricted in sequence space, otherwise site would already have been restricted
            if paired_site:
                primary_tmp_1[paired_site] = 'C'
                primary_tmp_2[paired_site] = 'G'

            # fold first tmp candidate
            folding_tmp_1 = _fold_primary(''.join(primary_tmp_1), self._env_config)

            # check first if hamming distance is at least the same for tmp candidate 1 (we do not want to increase hamming distance)
            if hamming_with_n(folding_tmp_1, self._target.dot_bracket) <= hamming_distance:
                # if hamming is ok, check if gc content would be better (abs diff)
                if self.gc_diff_abs_primary(''.join(primary_tmp_1)) < self.gc_diff_abs(self._design):
                    # if gc is better, asssign site to the design
                    self._design.assign_sites(0, site, self._target.get_paired_site(site))  # assign G/GC <=> action 0
            else:
                # do the same for candidate 2 if candidate 1 failed
                folding_tmp_2 = _fold_primary(''.join(primary_tmp_1), self._env_config)
                if hamming_with_n(folding_tmp_2, self._target.dot_bracket) <= hamming_distance:
                    if self.gc_diff_abs_primary(''.join(primary_tmp_2)) < self.gc_diff_abs(self._design):
                        self._design.assign_sites(3, site, self._target.get_paired_site(site))  # assign C/CG <=> action 3


    def _decrease_gc(self, site, paired_site=None, hamming_distance=0):
        """TODO
        """
        # create two temporary primaries to check for gc-improvement (two because we swap G/C with A and U)
        primary_tmp_1 = list(self._design.primary)
        primary_tmp_2 = list(self._design.primary)
        # if current site is A or U replace with G and C (and watson-crick pair base if site is paired)
        if self._design.primary[site] in ['G', 'C']:
            # assign G and C separately
            primary_tmp_1[site] = 'A'
            primary_tmp_2[site] = 'U'
            # and assign paired site if nucleotide is paired; paired site is not restricted in sequence space, otherwise site would already be restricted
            if paired_site:
                primary_tmp_1[paired_site] = 'U'
                primary_tmp_2[paired_site] = 'A'

            # fold first tmp candidate
            folding_tmp_1 = _fold_primary(''.join(primary_tmp_1), self._env_config)

            # check first if hamming distance is at least the same for tmp candidate 1 (we do not want to increase hamming distance)
            if hamming_with_n(folding_tmp_1, self._target.dot_bracket) <= hamming_distance:
                # if hamming is ok, check if gc content would be better (abs diff)
                if self.gc_diff_abs_primary(''.join(primary_tmp_1)) < self.gc_diff_abs(self._design):
                    # if gc is better, asssign site to the design
                    self._design.assign_sites(1, site, self._target.get_paired_site(site))  # assign A/AU <=> action 1
            else:
            # do the same for candidate 2 if candidate 1 failed
                folding_tmp_2 = _fold_primary(''.join(primary_tmp_1), self._env_config)
                if hamming_with_n(folding_tmp_2, self._target.dot_bracket) <= hamming_distance:
                    if self.gc_diff_abs_primary(''.join(primary_tmp_2)) < self.gc_diff_abs(self._design):
                        self._design.assign_sites(2, site, self._target.get_paired_site(site))  # assign U/UA <=> action 2


class _Target(object):
    """TODO
    Class of the target structure. Provides encodings and id.
    """

    _id_counter = 0

    def __init__(self, input, env_config):
        """TODO: rename dot_bracket input
        Initialize a target structure.

        Args:
             dot_bracket: dot_bracket encoded target structure.
             env_config: The environment configuration.
        """
        _Target._id_counter += 1
        self._env_config = env_config
        self.rng = np.random.default_rng(seed=env_config.seed)

        assert len(input) == 5, "Input does not contain enough information, require: (id, structure constraints, sequence constraints, GC content, desired energy)"

        self.id = input[0]
        # dot bracket may include extension symbols
        self.dot_bracket_X_int = input[1]

        # get all explicit X and O extensions and replace by X/O to get correct sequence constraints later
        self.X_plicits = {}

        # if there are any specific extensions given
        if re.search(r'X\d+|O\d+', self.dot_bracket_X_int):
            # loop until none are left
            while re.search(r'X\d+|O\d+', self.dot_bracket_X_int):
                # find first occurence of extension
                m = re.search(r'X\d+|O\d+', self.dot_bracket_X_int)
                # get the length of the extension
                r = int(m.group(0)[1:])
                # save the substitution and the index to insert later
                if self.dot_bracket_X_int[m.start()] == 'X':
                    sub = 'N' * r
                    self.X_plicits[m.start()] = sub
                    # replace the Xinteger match with X only to get correct sequence constraints
                    self.dot_bracket_X_int = re.sub('X\d+', 'X', self.dot_bracket_X_int, count=1)
                else:
                    sub = '.' * r
                    self.X_plicits[m.start()] = sub
                    # replace the Xinteger match with X only to get correct sequence constraints
                    self.dot_bracket_X_int = re.sub('O\d+', 'O', self.dot_bracket_X_int, count=1)

        # sequence constraints (if given) should not contain extension symbol
        seq = list(input[2]) if input[2] != '-' else []
        # gc content required if provided with file (TODO: set control_gc to true if provided and capture if not provided)
        self.gc = input[3] if not self._env_config.desired_gc else self._env_config.desired_gc
        # mfe currently not used
        self.mfe = input[4]
        # generate dot_bracket by removing the Xs and Os (dot bracket still contains int of enforced pairs)
        self.dot_bracket = self.dot_bracket_X_int.upper().replace('X', '').replace('O', '')
        # remove integers from dot_bracket_XO to find indices for variable length insertions
        self.dot_bracket_X = re.sub('\d+', '', self.dot_bracket_X_int)
        # but get all Xs (extend with N) and Os (extend with dots) to insert into sequence constraints
        Xs = [(i, self.dot_bracket_X[i]) for i in self.findall_X('X', self.dot_bracket_X.upper())]
        Os = [(i, self.dot_bracket_X[i]) for i in self.findall_X('O', self.dot_bracket_X.upper())]

        extensions = Xs + Os

        extensions = sorted(extensions, key=lambda x: x[0])

        from collections import defaultdict

        # start getting enforced pairs
        self.enforced_pairs = defaultdict(list)
        # if there are any ints left in the dot bracket (=enforced pairs; matching integer numbers correspond to matching brackets)
        if re.search(r'\d+', self.dot_bracket):
            tmp_db = list(self.dot_bracket)
            # while there are any ints left
            while re.search(r'\d+', self.dot_bracket):
                # find first occurence
                m = re.search(r'\d+', self.dot_bracket)
                # append them matching their int value
                self.enforced_pairs[m[0]].append(m.start() - 1)
                # but remove them from the list to get correct indices for further search
                del tmp_db[m.start():m.end()]
                # and also remove them from dot bracket
                self.dot_bracket = ''.join(tmp_db)

        # at the end check that all enforced pairs are balanced
        assert all([len(x) == 2 for x in self.enforced_pairs.values()]), "Found unbalanced explicit pairs"

        if not seq:
            self.sequence_constraints_X = None
            self.sequence_constraints = None
        else:
            self.sequence_constraints = ''.join(seq)

        if not self.sequence_constraints:
            sequence_constraints = ['N' for _ in self.dot_bracket]
            dot_bracket = [c for c in self.dot_bracket]
            for index, character in enumerate(self.dot_bracket):
                if character in ['A', 'C', 'G', 'U']:
                    sequence_constraints[index] = character
                    dot_bracket[index] = 'N'
            self.sequence_constraints = ''.join(sequence_constraints)
            self.dot_bracket = ''.join(dot_bracket)

        # if there are any extensions in the provided dot bracket
        for index, extension in extensions:
            # insert them into the sequence constraints
            seq[index:index] = extension
        self.sequence_constraints_X = ''.join(seq)
        # but remove them for sequence constraints only string
        self.sequence_constraints = self.sequence_constraints_X.upper().replace('O', '').replace('X', '')

        self._current_site = 0
        if not self.sequence_contains_iupac():

            self._partition = self.get_partition()
            self.encode_partial_pairing()
            self.local_target = self.assign_sequence_constraints()

            self.sequence_progress = self.local_target
            if self._env_config.encoding == 'tuple':
                self.padded_encoding = _encode_tuple(self.dot_bracket, self.sequence_constraints, self._env_config)
            elif self._env_config.encoding == 'multihot' and self._env_config.use_conv:
                self.padded_encoding = multihot(self.dot_bracket, self.sequence_constraints, self._env_config)
            else:
                self.padded_encoding = _encode_dot_bracket(self.local_target, self._env_config)

            if env_config.reward_function == 'structure_only':
                self.structure_parts_encoding = _encode_structure_parts(self.local_target)

        
    def sequence_contains_iupac(self):
        iupac = {
            'R': ['A', 'G'],
            'Y': ['C', 'U'],
            'S': ['G', 'C'],
            'W': ['A', 'U'],
            'K': ['G', 'U'],
            'M': ['A', 'C'],
            'B': ['C', 'G', 'U'],
            'D': ['A', 'G', 'U'],
            'H': ['A', 'C', 'U'],
            'V': ['A', 'C', 'G'],
        }
        for nuc in self.sequence_constraints:
            if nuc in iupac.keys():
                return True
        return False




    def sample_iupac(self, seq):
        if self._env_config.control_gc:
            p_GC = float(self.gc)
            p_AU = 1 - float(self.gc)
        else:
            p_GC = 0.5
            p_AU = 0.5
        

        iupac = {
            'R': [['A', 'G'], [p_AU, p_GC]],
            'Y': [['C', 'U'], [p_GC, p_AU]],
            'S': [['G', 'C'], [.5, .5]],
            'W': [['A', 'U'], [.5, .5]],
            'K': [['G', 'U'], [p_GC, p_AU]],
            'M': [['A', 'C'], [p_AU, p_GC]],
            'B': [['C', 'G', 'U'], [p_GC / 2, p_GC / 2, p_AU]],
            'D': [['A', 'G', 'U'], [p_AU / 2, p_GC, p_AU / 2]],
            'H': [['A', 'C', 'U'], [p_AU / 2, p_GC, p_AU / 2]],
            'V': [['A', 'C', 'G'], [p_AU, p_GC / 2, p_GC / 2]],
        }

        seq_copy = list(seq)
        # print(seq_copy)
        for i, sym in enumerate(seq):
            if sym in iupac.keys():
                # print(sym)
                # print(i)
                seq_copy[i] = np.random.choice(iupac[sym][0], p=iupac[sym][1])  # sampling without seed
                seq_copy[i] = self.rng.choice(iupac[sym][0], p=iupac[sym][1])  # sampling with seed
                # print(seq_copy[i])
        # print(''.join(seq_copy))
        return ''.join(seq_copy)


    def findall_X(self, pattern, s):
        """TODO
        """
        i = s.find(pattern)
        while i != -1:
            yield i
            i = s.find(pattern, i+1)

    def sample_random_target(self, min_length=0, max_length=0, min_dot_extension=0):
        """TODO
        """
        if not 'X' in self.dot_bracket_X and not 'O' in self.dot_bracket_X:  # no X to extend
            return _Target((self.id, self.dot_bracket_X, self.sample_iupac(self.sequence_constraints), self.gc, self.mfe), self._env_config)

        # make lists from dot-bracket inclusing Xs, Os, and enforced pairs to extend the strings
        db_list = list(self.dot_bracket_X_int)
        seq = list(self.sequence_constraints_X)

        # insert explicit length inserts
        x_plicit_extensions = []

        for index, value in enumerate(db_list):
            if value in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']:
                seq[index:index] = value

        for index, sub in self.X_plicits.items():
            index += int(np.sum(x_plicit_extensions))
            seq_sub = 'N' * len(sub)
            db_list[index:index] = sub
            seq[index:index] = seq_sub
            x_plicit_extensions.append(len(sub))

        seq = re.sub('\d+', '', ''.join(seq))
        seq = list(seq)

        xplicit_extensions_length = np.sum(x_plicit_extensions)

        num_Os = self.dot_bracket_X_int.upper().count('O')
        num_Xs_db = self.dot_bracket_X_int.upper().count('X')
        num_Xs_seq = self.sequence_constraints_X.upper().count('X')

        # insert minimum dots at each extension site (Os)
        min_O = ["."] * min_dot_extension
        min_O_N = ["N"] * min_dot_extension

        for i in range(num_Os):
            all_O = [i for i in self.findall_X('O', ''.join(db_list))]
            all_O_seq = [i for i in self.findall_X('O', ''.join(seq))]
            db_list[all_O[i]:all_O[i]] = min_O
            seq[all_O_seq[i]:all_O_seq[i]] = min_O_N


        # compute borders for the extension with Ns
        max_extension = max_length - (num_Os * min_dot_extension) - len(self.dot_bracket) - xplicit_extensions_length
        min_extension = min_length - (len(self.dot_bracket) + (num_Os * min_dot_extension) + xplicit_extensions_length)

        # if minimal length is already covered by current dot bracket length and minimum extensions by dots, set lower bound to zero
        if min_extension < 0:
            min_extension = 0

        if max_extension <= 0:
            # print("Reached extension limit before extension of Xs.")
            dot_bracket = (self.id, ''.join(db_list), self.sample_iupac(''.join(seq).upper().replace('X', '').replace('O', '')), self.gc, self.mfe)
            return _Target(dot_bracket, self._env_config)

        if 'X' in self.dot_bracket_X:
            # select random insert length for Xs
            # random_insert_length = np.random.randint(min_extension, max_extension)  # sampling without seed
            random_insert_length = self.rng.integers(min_extension, max_extension, endpoint=True)  # sampling with seed

            # and create list of Ns of length random_insert length to insert into dot_bracket
            substitutions_N = ["N"] * random_insert_length

            # iterate over the possible inserts
            while substitutions_N:
                # find all extension positions in the dot_bracket
                all_X = [i for i in self.findall_X('X', ''.join(db_list))]
                all_X_seq = [i for i in self.findall_X('X', ''.join(seq))]

                # get an insertion position
                # insertion_index = np.random.randint(0, len(all_X))  # sampling without seed
                insertion_index = self.rng.integers(0, len(all_X))  # sampling with seed
                insertion_position_db = all_X[insertion_index]
                insertion_position_seq = all_X_seq[insertion_index]

                # get a length of the substitution to insert
                # substitution_len = np.random.randint(0, len(substitutions_N) + 1)  # sampling without a seed
                substitution_len = self.rng.integers(0, len(substitutions_N), endpoint=True)  # sampling with seed

                # get the substitution from the list of Ns
                # substitution = np.random.choice(substitutions_N, substitution_len)  # sampling without seed
                substitution = self.rng.choice(substitutions_N, substitution_len)  # sampling without seed

                # substitute in seq and db and update list of possible inserts
                db_list[insertion_position_db:insertion_position_db] = substitution
                seq[insertion_position_seq:insertion_position_seq] = substitution

                substitutions_N = substitutions_N[substitution_len:]

        if not 'O' in self.dot_bracket_X:
            dot_bracket = (self.id, ''.join(db_list), self.sample_iupac(''.join(seq).upper().replace('X', '').replace('O', '')), self.gc, self.mfe)
            return _Target(dot_bracket, self._env_config)


        # get current dot-bracket to calculate the remaining maximum insertion length
        tmp_db = re.sub('\d+', '', ''.join(db_list).replace('X', '').replace('O', ''))

        # Calculate min and max extensions from the given length and already placed insertions...
        min_O_extension = min_length - len(tmp_db)
        max_O_extension = max_length - len(tmp_db)

        if min_O_extension < 0:
            min_O_extension = 0

        if max_extension <= 0:
            # print("Reached extension limit before extension of Os.")
            dot_bracket = (self.id, ''.join(db_list), self.sample_iupac(''.join(seq).upper().replace('X', '').replace('O', '')), self.gc, self.mfe)
            return _Target(dot_bracket, self._env_config)

        # select random insert length for Os
        # random_O_insert_length = np.random.randint(min_O_extension, max_O_extension)  # sampling without seed
        random_O_insert_length = self.rng.integers(min_O_extension, max_O_extension, endpoint=True)  # sampling with seed

        # and create list of dots of length random_O_insert_length to insert into dot_bracket
        substitutions_O = ["."] * random_O_insert_length
        substitutions_O_seq = ["N"] * random_O_insert_length

        # iterate over the possible inserts
        while substitutions_O:
            # update indices
            all_O = [i for i in self.findall_X('O', ''.join(db_list))]
            all_O_seq = [i for i in self.findall_X('O', ''.join(seq))]

            # get an insertion position
            # insertion_index = np.random.randint(0, len(all_O))  # sampling without seed
            insertion_index = self.rng.integers(0, len(all_O))  # sampling with seed
            insertion_position_db = all_O[insertion_index]
            insertion_position_seq = all_O_seq[insertion_index]

            # get a length of the substitution to insert
            # substitution_len = np.random.randint(0, len(substitutions_O) + 1)  # sampling without seed
            substitution_len = self.rng.integers(0, len(substitutions_O), endpoint=True)  # sampling with seed

            # get the substitution from the list of Ns
            # substitution = np.random.choice(substitutions_O, substitution_len)  # sampling without seed
            substitution = self.rng.choice(substitutions_O, substitution_len)  # sampling with seed
            # substitution_seq = np.random.choice(substitutions_O_seq, substitution_len)  # sampling without seed
            substitution_seq = self.rng.choice(substitutions_O_seq, substitution_len)  # sampling with seed

            # substitute in seq and db and update list of possible inserts
            db_list[insertion_position_db:insertion_position_db] = substitution
            seq[insertion_position_seq:insertion_position_seq] = substitution_seq
            substitutions_O = substitutions_O[substitution_len:]

        # create new input for new target
        dot_bracket = (self.id, ''.join(db_list), self.sample_iupac(''.join(seq).upper().replace('X', '').replace('O', '')), self.gc, self.mfe)

        # and return the target
        return _Target(dot_bracket, self._env_config)

    def encode_partial_pairing(self):
        """TODO
        """
        self._pairing_encoding = [None] * len(self.dot_bracket)

        for start, end in self.partition[1]:
            encoding = _encode_pairing(self.dot_bracket[start:end])
            for index, site in enumerate(encoding):
                if site is not None:
                    self._pairing_encoding[index + start] = site + start

        for enforced_pair in self.enforced_pairs.values():
            self._pairing_encoding[enforced_pair[0]] = enforced_pair[1]
            self._pairing_encoding[enforced_pair[1]] = enforced_pair[0]

    def tmp_generate_sequence_constraints(self):
        sequence_constraints = []
        for site in self.local_target:
            if site in ['A', 'C', 'G', 'U']:
                sequence_constraints.append(site)
            else:
                sequence_constraints.append('N')
        return ''.join(sequence_constraints)

    def assign_sequence_constraints(self):
        """
        Generate/Update the local target representation.
        """
        pair_assignments = {'A': 'U', 'U': 'A', 'G': 'C', 'C': 'G'}

        # get all structure constraints to be partially overwritten by sequence constraints
        new_local_target = [site for site in self.dot_bracket.rstrip()]
        # loop over sequence constraints
        for index, site in enumerate(self.sequence_constraints):
            # if a site is constrained in the sequence space
            if site != 'N':
                # assign it to the local target
                paired_site = self._pairing_encoding[index]
                # and get the paired site if exists
                new_local_target[index] = site
                # if there is a paired site, assign it, too
                if paired_site:
                    # but be careful not to overwrite existing sequence constraints that might differ from watson-crick scheme
                    if not self.sequence_constraints[paired_site] in ['A', 'C', 'G', 'U']:
                        new_local_target[paired_site] = pair_assignments[site]
        return ''.join(new_local_target)

    def get_partition(self):
        """
        Partition the inputs into sequence and structure parts (removing Ns).
        """
        sequence_pattern = re.compile(r"[A, C, G, U]+")
        structure_pattern = re.compile(r"[0, 1, 2]+")

        sequence_parts = []
        structure_parts = []

        current_index = 0

        for pattern in sequence_pattern.findall(self.sequence_constraints):
            start, end = re.search(pattern, self.sequence_constraints[current_index:]).span()
            sequence_parts.append((start + current_index, end + current_index))
            current_index += end

        current_index = 0

        tmp_db = [x for x in self.dot_bracket.rstrip()]

        for index, site in enumerate(self.dot_bracket.rstrip()):
            if site not in ['N', 'A', 'C', 'G', 'U']:  # ACGU is quite useless, isn't it? dot_bracket should not contain sequence parts?
                if site == '.':
                    tmp_db[index] = '0'
                elif site == '(':
                    tmp_db[index] = '1'
                else:
                    tmp_db[index] = '2'

        tmp_db = ''.join(tmp_db)

        for pattern in structure_pattern.findall(tmp_db):
            start, end = re.search(pattern, tmp_db[current_index:]).span()
            structure_parts.append((start + current_index, end + current_index))
            current_index += end

        return sequence_parts, structure_parts

    def __len__(self):
        return len(self.dot_bracket)

    def get_paired_site(self, site):
        """
        Get the paired site for <site> (base pair).

        Args:
            site: The site to check the pairing site for.

        Returns:
            The site that pairs with <site> if exists.TODO
        """
        return self._pairing_encoding[site]

    def reset(self):
        """
        Reset the constraints after a complete episode.
        Required for sequence_progress state since we assign nucleotides during the episode.
        """
        # reset the constraints
        self.sequence_progress = self.local_target if self._env_config.local_design else self.dot_bracket
        # and the encoding
        if self._env_config.encoding == "multihot" and self._env_config.use_conv:
            self.padded_encoding = multihot(self.dot_bracket, self.sequence_constraints, self._env_config)
        elif self._env_config.encoding == "tuple":
            self.padded_encoding = _encode_tuple(self.dot_bracket, self.sequence_constraints, self._env_config)
        else:
            self.padded_encoding = _encode_dot_bracket(self.dot_bracket, self._env_config) if not self._env_config.local_design else _encode_dot_bracket(self.local_target, self._env_config)

    def assign_sites(self, index, value, paired_site):
        """
        Assign sites during the episode if state is sequence_progress.

        """
        new_local_target = list(self.sequence_progress)
        if paired_site:
            new_local_target[index[0]] = value[0]
            new_local_target[index[1]] = value[1]
        else:
            new_local_target[index] = value
        self.sequence_progress = ''.join(new_local_target)
        if self._env_config.encoding == "multihot" and self._env_config.use_conv:
            self.padded_encoding = multihot(self.dot_bracket, self.sequence_progress, self._env_config)
            # print(self.padded_encoding)
        elif self._env_config.encoding == "tuple":
            self.padded_encoding = _encode_tuple(self.dot_bracket, self.sequence_constraints, self._env_config)
        else:
            self.padded_encoding = _encode_dot_bracket(self.sequence_progress, self._env_config)


    def reset_counter(self):
        self._current_site = 0
        self.structure_parts_encoding = _encode_structure_parts(self.local_target)

    @property
    def partition(self):
        return self._partition

    @property
    def next_structure_site(self):
        try:
            while self.structure_parts_encoding[self._current_site] is None:
                self._current_site += 1
            self.structure_parts_encoding[self._current_site] = None
            if self._env_config.predict_pairs and self.get_paired_site(self._current_site):
                self.structure_parts_encoding[self.get_paired_site(self._current_site)] = None
            return self._current_site
        except IndexError:
            return None

    @property
    def current_site(self):
        return self._current_site


class _Design(object):
    """
    Class of the designed candidate solution.
    """

    action_to_base = {0: "G", 1: "A", 2: "U", 3: "C", 4:'_'}
    action_to_pair = {0: "GC", 1: "AU", 2: "UA", 3: "CG"}

    def __init__(self, length=None, primary=None):
        """
        Initialize a candidate solution.

        Args:
            length: The length of the candidate solution.
            primary: The sequence of the candidate solution.
        """
        if primary:
            self._primary_list = primary
        else:
            self._primary_list = [None] * length
        self._dot_bracket = None
        self._current_site = 0
        self._last_assignment = None

    def get_mutated(self, mutations, sites):
        """
        Locally change the candidate solution.

        Args:
            mutations: Possible mutations for the specified sites
            sites: The sites to be mutated

        Returns:
            A Design object with the mutated candidate solution.
        """
        mutatedprimary = self._primary_list.copy()
        for site, mutation in zip(sites, mutations):
            mutatedprimary[site] = mutation
        return _Design(primary=mutatedprimary)

    def assign_sites(self, action, site, paired_site=None, predict_pairs=False):
        """
        Assign nucleotides to sites for designing a candidate solution.

        Args:
            action: The agents action to assign a nucleotide.
            site: The site to which the nucleotide is assigned to.
            paired_site: defines if the site has a corresponding pairing partner or not.
        """
        
        self._current_site += 1
        
        if predict_pairs and paired_site:
            base_current, base_paired = self.action_to_pair[action]
            self._primary_list[site] = base_current
            self._primary_list[paired_site] = base_paired
            self._last_assignment = ((site, paired_site), (base_current, base_paired), True)
        else:
            self._primary_list[site] = self.action_to_base[action]
            self._last_assignment = (site, self.action_to_base[action], False)



    @property
    def last_assignment(self):
        return self._last_assignment

    @property
    def first_unassigned_site(self):
        try:
            while self._primary_list[self._current_site] is not None:
                self._current_site += 1
            return self._current_site
        except IndexError:
            return None

    @property
    def primary(self):
        return "".join(self._primary_list)


def _random_epoch_gen(data):
    """
    Generator to get epoch data.

    Args:
        data: The targets of the epoch
    """
    while True:
        for i in np.random.permutation(len(data)):
            yield data[i]

def _sorted_data_gen(data):
    data = sorted(data, key=lambda x: len(x))
    while True:
        for target in data:
            yield target


def hamming_with_n(s1, s2):
    distance = 0
    for c1, c2 in zip(s1, s2):
        if c1 != c2:
            if c1 == 'N' or c2 == 'N':
                continue
            elif c1 == '|' and c2 in ['(', ')', '[', ']', '{', '}', '<', '>']:
                continue
            elif c2 == '|' and c1 in ['(', ')', '[', ']', '{', '}', '<', '>']:
                continue
            else:
                distance += 1
    return distance

def change_encoding(structure):
    structure = ["(" if x == "(0" else x for x in structure]
    structure = [")" if x == ")0" else x for x in structure]
    structure = ["[" if x == "(1" else x for x in structure]
    structure = ["]" if x == ")1" else x for x in structure]
    structure = ["{" if x == "(2" else x for x in structure]
    structure = ["}" if x == ")2" else x for x in structure]
    structure = ["<" if x == "(3" else x for x in structure]
    structure = [">" if x == ")3" else x for x in structure]
    return structure


@dataclass
class EpisodeInfo:
    """
    Information class.
    """

    __slots__ = ["target_id", "time", "normalized_hamming_distance", "gc_content", "agent_gc", "desired_gc", "delta_gc", "gc_satisfied", "folding", "folding_counter", "candidate", "interactions", "target"]
    target_id: int
    time: float
    normalized_hamming_distance: float
    folding: str
    folding_counter: int
    candidate: str
    gc_content: float
    agent_gc: float
    desired_gc: float
    delta_gc: float
    gc_satisfied: bool
    interactions: int
    target: int

def stockholm2records(sto_file):
    align = AlignIO.read(sto_file, "stockholm")
    return align


def stockholm2df(sto_file):
    align = stockholm2records(sto_file)
    d_list = [{'Id': r.name, 'sequence': r.seq} for r in align]
    return pd.DataFrame(d_list)

def stockholm2idlist(sto_file):
    align = stockholm2records(sto_file)
    return [r.name for r in align]

def read_e_values(tab_file):
    with open(tab_file) as f:
        lines = f.readlines()
    header = [x.strip().replace('#', '') for x in (lines[0].split('\t'))]
    data = []
    for line in lines[1:]:
        line = line.strip()
        if line.startswith('# Program:'):
            continue
        elif line.startswith('# Version:'):
            continue
        elif line.startswith('# Pipeline mode:'):
            continue
        elif line.startswith('# Query file:'):
            continue
        elif line.startswith('# Target file:'):
            continue
        elif line.startswith('# Option settings:'):
            continue
        elif line.startswith('# Current dir:'):
            continue
        elif line.startswith('# Date:'):
            continue
        elif line.startswith('# [ok]'):
            continue
        elif line.startswith('#----'):
            continue
        else:
            line_list = [x.replace('#', '') for x in line.split()]
            if line_list and line_list[0] != '':
                data.append(line_list)
    # print(data)
    # print(header)
    if data:
        df = pd.DataFrame(data, columns=['target name', 'accession', 'query name', 'accession', 'mdl', 'mdl from', 'mdl to', 'seq from', 'seq to', 'strand', 'trunc', 'pass', 'gc', 'bias', 'score', 'E-value', 'inc', 'description of target'])
        return df
    else:
        return pd.DataFrame(columns=['target name', 'accession', 'query name', 'accession', 'mdl', 'mdl from', 'mdl to', 'seq from', 'seq to', 'strand', 'trunc', 'pass', 'gc', 'bias', 'score', 'E-value', 'inc', 'description of target'])


class Infernal():
    def __init__(self,
                 working_dir: str = 'working_dir',
                 E: float = 0.001,
                 incE: float = 0.001,
                 aln_outfile = 'infernal.tab',
                 ):
        self.working_dir = str(Path(working_dir).resolve())
        self.aln = str(Path(working_dir, aln_outfile).resolve())
        self.E = E
        self.incE = incE
        # print(self.E, self.incE)

    def search_database(self,
                        cm_database: str,
                        identifier: str,
                        fasta_db: str,
                        ) -> list:

        call1 = ["echo", str(identifier)]
        call2 = ["cmfetch", "-f", f"{cm_database}", "-"]
        call3 = ["cmsearch", "--tblout", f"{self.aln}", "--nohmmonly", "-E", f"{self.E}", "-", f"{fasta_db}"]  # "-A", f"{self.aln}", "--incE", f"{self.incE}",  "-T", "30.00", "-Z", "742849.287494",

        ps1 = subprocess.Popen(call1, stdout=subprocess.PIPE)
        ps2 = subprocess.Popen(call2, stdin=ps1.stdout, stdout=subprocess.PIPE)
        subprocess.run(call3, stdin=ps2.stdout, stdout=subprocess.DEVNULL)

        # hit_ids = stockholm2idlist(self.aln)
        hits = read_e_values(self.aln)

        return hits

    def get_family_information(self,
                               queries_fasta_path : str,
                               cm_path : str,
                               clanin_path : str,
                               outpath : str,
                               ):
        # subprocess.call(["cmpress", cm_path])
        subprocess.call(["cmscan", "--rfam", "--cut_ga", "--nohmmonly", "--oskip", "--tblout", outpath, "--fmt", "2", "--clanin", clanin_path, cm_path, queries_fasta_path])



    def search(self, cm_path: str, fasta_path: str, outfile: str = None):
        if outfile:
            subprocess.call(["cmsearch", "-A", f"{outfile}", "-E", f"{self.E}", "--incE", f"{self.incE}", cm_path, fasta_path])
            try:
                hit_ids = stockholm2idlist(outfile)
            except ValueError as e:
                print('Note:', 'No hits satisfy inclusion thresholds; no alignment saved')
                hit_ids = []
            except FileNotFoundError as e:
                print('Note:', 'no alignment file found')
                hit_ids = []
        else:
            subprocess.call(["cmsearch", "-A", f"{self.aln}", "-E", f"{self.E}", "--incE", f"{self.incE}", cm_path, fasta_path])
            try:
                hit_ids = stockholm2idlist(self.aln)
            except ValueError as e:
                print('Note:', 'No hits satisfy inclusion thresholds; no alignment saved')
                hit_ids = []
            except FileNotFoundError as e:
                print('Note:', 'no alignment file found')
                hit_ids = []

        return hit_ids

    def build_noss(self, stk_path: str, out_stem: str, name: str = '', cm_dir: str = ''):
        if not cm_dir:
            if not name:
                subprocess.call(["cmbuild", "--noss", f"{self.working_dir}/{out_stem}.cm", stk_path])
            else:
                subprocess.call(["cmbuild", "--noss", "-F", "-n", name, f"{self.working_dir}/{out_stem}.cm", stk_path])
        else:
            if not name:
                subprocess.call(["cmbuild", "--noss", f"{cm_dir}/{out_stem}.cm", stk_path])
            else:
                subprocess.call(["cmbuild", "--noss", "-F", "-n", name, f"{cm_dir}/{out_stem}.cm", stk_path])

    def build(self, stk_path: str, out_stem: str, name: str = '', cm_dir: str = ''):
        if not cm_dir:
            if not name:
                subprocess.call(["cmbuild", "-F", f"{self.working_dir}/{out_stem}.cm", stk_path])
            else:
                subprocess.call(["cmbuild", "-F", "-n", name, f"{self.working_dir}/{out_stem}.cm", stk_path])
        else:
            if not name:
                subprocess.call(["cmbuild", "-F", f"{cm_dir}/{out_stem}.cm", stk_path])
            else:
                subprocess.call(["cmbuild", "-F", "-n", name, f"{cm_dir}/{out_stem}.cm", stk_path])


    def calibrate(self, cm_path: str):
        subprocess.call(["cmcalibrate", cm_path])

    def sample_msa(self, cm_path : str, outpath : str, n : int = 1000, seed : int = 42):
        subprocess.call(["cmemit", "-o", outpath, "-a", "-N", str(n), "--seed", str(seed), cm_path])


class RnaDesignEnvironment(Environment):
    """
    The environment for RNA design using deep reinforcement learning.
    """

    def __init__(self, dot_brackets, env_config):
        """TODO
        Initialize an environemnt.

        Args:
            env_config: The configuration of the environment.
        """
        self._env_config = env_config

        # targets = [_Target(dot_bracket, self._env_config) for dot_bracket in dot_brackets]
        targets = [_Target((i, db, seq, gc, en), self._env_config) for i, db, seq, gc, en in dot_brackets]
        # print(targets)
        self._target_gen = _random_epoch_gen(targets)
        if self._env_config.data_type == 'random-sort':
            self._target_gen = _sorted_data_gen(targets)

        self.target = None
        self.design = None
        self._folding = None
        self.episodes_info = []

        # try CM hitting as task
        self.cm_design = env_config.cm_design
        if self.cm_design:
            self.working_dir = env_config.working_dir
            self.max_e = 100
            Path(self.working_dir).mkdir(parents=True, exist_ok=True)
            self.cm_name = env_config.cm_name  # 'RF00651'  # 'RF00162'  # SAM riboswitch; LEN 108, MaxL 270;  RF00651 -> mir-221 CLEN 85
            self.cm_db = env_config.cm_path
            self.infernal = Infernal(working_dir=self.working_dir,
                        E=self.max_e,
                        incE=self.max_e)
        
        self.rri_design = env_config.rri_design
        if self.rri_design:
            self.rri_target = env_config.rri_target

        self.alpha = 1.0
        self.beta = 1.0
        # self.alpha_beta_threshold = 0.02
        # self.gc_look_back = -50

    def __str__(self):
        return "RnaDesignEnvironment"

    def seed(self, seed):
        return None

    def reset(self):
        """
        Reset the environment. First function called by runner. Returns first state.

        Returns:
            The first state.
        """
        self._env_config.interactions += 1
        self.target = next(self._target_gen)

        if not self._env_config.control_gc:
            self._env_config.control_gc = self.target.gc is not None

        # print(self.target.sequence_constraints)

        # get a random target with random extension of Os and Xs in the structure constraints
        # if self._env_config.variable_length:
        self.target = self.target.sample_random_target(self._env_config.min_length, self._env_config.max_length, self._env_config.min_dot_extension)
        # print(self.target.sequence_constraints)

        # inject desired gc-content if provided with env-config, else use gc content provided with file
        if self._env_config.control_gc and self._env_config.desired_gc:
            self.target.gc = self._env_config.desired_gc

        self.target.reset()

        if self._env_config.reward_function == 'structure_only':
            self.target.reset_counter()

        self.design = _Design(len(self.target))

        return self._get_state()

    def _apply_action(self, action):
        """
        Assign a nucleotide to a site.

        Args:
            action: The action chosen by the agent.
        """
        current_site = self.design.first_unassigned_site if not self._env_config.reward_function == 'structure_only' else self.target.current_site
        paired_site = self.target.get_paired_site(current_site) if self._env_config.predict_pairs else None  # None for unpaired sites
        self.design.assign_sites(action, current_site, paired_site, self._env_config.predict_pairs)

    def _get_state(self):
        """
        Get a state dependend on the padded encoding of the target structure.

        Returns:
            The next state.
        """
        start = self.target.next_structure_site if self._env_config.reward_function == 'structure_only' else self.design.first_unassigned_site

        if not self._env_config.state_representation == 'n-gram':
            if self.design.last_assignment:
                index, value, paired_site = self.design.last_assignment
                self.target.assign_sites(index, value, paired_site)

        if self._env_config.encoding == "multihot" and self._env_config.use_conv:
            state = self.target.padded_encoding[start : start + 2 * self._env_config.state_radius + 1, :, :]
            return state

        return self.target.padded_encoding[
            start : start + 2 * self._env_config.state_radius + 1
        ]

    def _get_local_design_loss(self, design):
        # initialize distance
        distance = 0
        # get folding if structure and sequence are evaluated
        folding = _fold_primary(design.primary, self._env_config) if self._env_config.reward_function == 'sequence_and_structure' else None
        #
        sequence_parts, folding_parts = self.target.partition

        if self._env_config.reward_function == 'sequence_and_structure':
            distance += hamming_with_n(design.primary, self.target.sequence_constraints)
        else:
            design = [c for c in design._primary_list]

            for index, site in enumerate(self.target.local_target):
                if site in ['A', 'C', 'G', 'U']:
                    design[index] = site

            self.design = _Design(primary=[c for c in ''.join(design).rstrip()])

            folding = _fold_primary(self.design.primary, self._env_config)
            # print(folding)

        distance += hamming_with_n(folding, self.target.dot_bracket)

        return distance, folding


    def _local_improvement(self, folded_design):
        """
        Compute Hamming distance of locally improved candidate solutions.
        Returns:
            The minimum Hamming distance of all imporved candidate solutions.
        """
        def flatten_list(list_):
            return [item for sublist in list_ for item in sublist]
        # print("enter LIS")

        if self._env_config.local_design:
            sequence_parts, structure_parts = self.target.partition
            differing_sites_per_part = []

            if self._env_config.reward_function == 'sequence_and_structure':
                for start, end in sequence_parts:
                    string_difference = _string_difference_indices(
                        self.target.sequence_constraints[start:end], self.design.primary[start:end]
                    )
                    string_difference = [index + start for index in string_difference]
                    differing_sites_per_part.append(string_difference)

            for start, end in structure_parts:
                string_difference = _string_difference_indices(
                    self.target.dot_bracket[start:end], self._folding[start:end]
                )
                string_difference = [index + start for index in string_difference]
                differing_sites_per_part.append(string_difference)

            differing_sites = flatten_list(differing_sites_per_part)
        else:

            differing_sites = _string_difference_indices(
                self.target.dot_bracket, folded_design
            )
        differing_copy = differing_sites.copy()
        for index in differing_sites:
            # print(index)
            # print(self.target.sequence_constraints[index])
            if self.target.sequence_constraints[index] != 'N':
                # print(differing_sites)
                # print(index)
                # print(self.target.sequence_constraints[index])
                differing_copy.remove(index)
                # print(differing_sites)
        differing_sites = differing_copy
        # print(differing_sites)
        hamming_distances = []
        for mutation in product("AGCU", repeat=len(differing_sites)):
            mutated = self.design.get_mutated(mutation, differing_sites)
            if self._env_config.local_design:
                hamming_distance, _ = self._get_local_design_loss(mutated)
            else:
                folded_mutated = _fold_primary(mutated.primary, self._env_config)
                hamming_distance = hamming_with_n(folded_mutated, self.target.dot_bracket)
            hamming_distances.append((hamming_distance, mutated))
            if hamming_distance == 0:  # For better timing results
                # print(f"in LIS {hamming_with_n(mutated.primary, self.target.sequence_constraints)}")
                # print(_fold_primary(mutated.primary, self._env_config))
                # print(mutated.primary)
                # print(f"in LIS: {self._get_local_design_loss(mutated)}")
                return (0, mutated)
        return min(hamming_distances, key=lambda x: x[0])

    def reward_local_design(self):

        self.gc_controler = None

        distance, self._folding = self._get_local_design_loss(self.design)
        gc_initial = (self.design.primary.upper().count('G') + self.design.primary.upper().count('C')) / len(self.design.primary)
        # print(self._folding)
        # print(f"initial: {(distance, self._folding)}")
        # print(self.design.primary)
        self.gc_controler = GcControler(self.design, self.target, self._env_config.gc_tolerance, self._env_config)
        # print(f"Before LIS {hamming_with_n(self.design.primary, self.target.sequence_constraints)}")

        # distance might get larger than sites in target because of overlapping constraints when algorithm also places sequence
        if distance > len(self.target):
            distance = len(self.target)

        if 0 < distance < self._env_config.mutation_threshold:
            distance, self.design = self._local_improvement(self._folding)
            # print(self._env_config.folding_counter)
            # print(f"after LIS {self._get_local_design_loss(self.design)}")
            # self._folding = _fold_primary(self.design.primary, self._env_config)
            # print(f"After LIS {hamming_with_n(self.design.primary, self.target.sequence_constraints)}")
            if self._env_config.control_gc:
                # print('enter GC-control')

                self.gc_controler = GcControler(self.design, self.target, self._env_config.gc_tolerance, self._env_config)

                self.design = self.gc_controler.gc_improvement_step(hamming_distance=distance)

                self._folding = _fold_primary(self.design.primary, self._env_config)

                distance = hamming_with_n(self._folding, self.target.dot_bracket)
                gc_diff_abs = self.gc_controler.gc_diff_abs(self.design)

                if self.gc_controler.gc_satisfied(self.design):
                    gc_penalty = 0
                else:
                    gc_penalty = gc_diff_abs
                # if not len(self.episodes_info) < self.gc_look_back:
                #     if np.mean([abs(ei.gc_content - self.gc_controler._desired_gc)  for ei in self.episodes_info[self.gc_look_back:]]) < self.alpha_beta_threshold:
                #         self.alpha = min(1.0, self.alpha + 0.02)
                        # self.beta = max(0.0, self.beta - 0.02)
                    # else:
                    #     self.alpha = max(0.0, self.alpha - 0.02)
                    #     self.beta = min(1.0, self.beta + 0.02)

                normalized_distance = (distance / len(self.target))

                general_distance = self.alpha * normalized_distance + self.beta * gc_penalty

                if (general_distance) > 1.0:
                    general_distance = 1.0

                episode_info = EpisodeInfo(
                target_id=self.target.id,
                time=time.time(),
                normalized_hamming_distance=general_distance,
                folding=self._folding,
                folding_counter=self._env_config.folding_counter,
                candidate=self.design.primary,
                agent_gc=gc_initial,
                desired_gc=self.target.gc,
                gc_content=(self.design.primary.upper().count('G') + self.design.primary.upper().count('C')) / len(self.design.primary),
                delta_gc=gc_diff_abs,
                gc_satisfied=self.gc_controler.gc_satisfied(self.design),
                interactions=self._env_config.interactions,
                target=self.target.dot_bracket,
                )
                self.episodes_info.append(episode_info)

                return (1 - (general_distance)) ** self._env_config.reward_exponent

        if not self._env_config.control_gc:
            # print(self._env_config.folding_counter)
            # print(f"final rewarding function: {self._get_local_design_loss(self.design)}")
            self._folding = _fold_primary(self.design.primary, self._env_config)
            normalized_distance = (distance / len(self.target))  # is this buggy?
            # print(normalized_distance)

            episode_info = EpisodeInfo(
            target_id=self.target.id,
            time=time.time(),
            normalized_hamming_distance=normalized_distance,
            folding=self._folding,
            folding_counter=self._env_config.folding_counter,
            candidate=self.design.primary,
            agent_gc=gc_initial,
            desired_gc=self.target.gc,
            gc_content=(self.design.primary.upper().count('G') + self.design.primary.upper().count('C')) / len(self.design.primary),
            delta_gc=None,
            gc_satisfied=self.gc_controler.gc_satisfied(self.design),
            interactions=self._env_config.interactions,
            target=self.target.dot_bracket,
            )
            self.episodes_info.append(episode_info)

            return (1 - normalized_distance) ** self._env_config.reward_exponent
        else:
            gc_diff_abs = self.gc_controler.gc_diff_abs(self.design)

            if self.gc_controler.gc_satisfied(self.design):
                gc_penalty = 0
            else:
                gc_penalty = gc_diff_abs

            normalized_distance = (distance / len(self.target))
            # if not len(self.episodes_info) < self.gc_look_back:
            #     if np.mean([abs(ei.gc_content - self.gc_controler._desired_gc)  for ei in self.episodes_info[self.gc_look_back:]]) < self.alpha_beta_threshold:
            #         self.alpha = min(1.0, self.alpha + 0.02)
                    # self.beta = max(0, self.beta - 0.02)
                # else:
                #     self.alpha = max(0, self.alpha - 0.02)
                #     self.beta = min(1.0, self.beta + 0.02)

            general_distance = self.alpha * normalized_distance + self.beta * gc_penalty

            if (general_distance) > 1.0:
                general_distance = 1.0

            episode_info = EpisodeInfo(
            target_id=self.target.id,
            time=time.time(),
            normalized_hamming_distance=general_distance,
            folding=self._folding,
            folding_counter=self._env_config.folding_counter,
            candidate=self.design.primary,
            agent_gc=gc_initial,
            desired_gc=self.target.gc,
            gc_content=(self.design.primary.upper().count('G') + self.design.primary.upper().count('C')) / len(self.design.primary),
            delta_gc=gc_diff_abs,
            gc_satisfied=self.gc_controler.gc_satisfied(self.design),
            interactions=self._env_config.interactions,
            target=self.target.dot_bracket,
            )
            self.episodes_info.append(episode_info)

            return (1 - (general_distance)) ** self._env_config.reward_exponent

    def _reward_cm_design(self):
        # self.design._primary_list = [x if x is not None else y for x, y in zip(self.design._primary_list, self.target.sequence_constraints)]
        self.design._primary_list = [x if x is not None else y for x, y in zip(self.design._primary_list, self.target.local_target)]
        fasta_path = Path(self.working_dir, 'tmp.fasta')

        with open(f"{fasta_path.resolve()}", 'w') as f:
            f.write(f">{self.target.id}\n{''.join(self.design.primary)}")
        
        

        try:
            hit_df = self.infernal.search_database(cm_database=self.cm_db, identifier=self.cm_name, fasta_db=str(fasta_path.resolve()))
        except ValueError as e:
            print(e)
        except Exception as e:
            print(e)
        if hit_df.empty:
            reward = -200
        else:
            hit_df.loc[:, 'score'] = hit_df['score'].astype(float)
            reward = hit_df['score'].max()
        fasta_path.unlink()

        episode_info = EpisodeInfo(
            target_id=self.target.id,
            time=time.time(),
            normalized_hamming_distance=reward,
            folding=None,
            folding_counter=1,
            candidate=self.design.primary,
            agent_gc=None,
            desired_gc=None,
            gc_content=None,
            delta_gc=None,
            gc_satisfied=None,
            interactions=None,
            target=self.target.dot_bracket,
            )
        self.episodes_info.append(episode_info)

        return reward

    def _reward_rri_design(self):
        # self.design._primary_list = [x if x is not None else y for x, y in zip(self.design._primary_list, self.target.sequence_constraints)]
        self.design._primary_list = [x if x is not None else y for x, y in zip(self.design._primary_list, self.target.local_target)]

        design = ''.join(self.design.primary)

        output = subprocess.check_output(['IntaRNA', self.rri_target, design]) 

        if output.decode('utf-8').split('\n')[-3].startswith('energy: '):
            energy = float(output.decode('utf-8').split('\n')[-3].replace('energy: ', '').replace(' kcal/mol', ''))
            reward = -1 * energy
        else:
            reward = 0

        reward = reward ** 3

        reward = max(0, reward)
        
        episode_info = EpisodeInfo(
        target_id=self.target.id,
        time=time.time(),
        normalized_hamming_distance=reward,
        folding=None,
        folding_counter=1,
        candidate=design,
        agent_gc=None,
        desired_gc=None,
        gc_content=None,
        delta_gc=None,
        gc_satisfied=None,
        interactions=None,
        target=self.target.dot_bracket,
        )
        
        self.episodes_info.append(episode_info)

        return reward

    def _get_reward(self, terminal):
        """
        Compute the reward after assignment of all nucleotides.

        Args:
            terminal: Bool defining if final timestep is reached yet.

        Returns:
            The reward at the terminal timestep or 0 if not at the terminal timestep.
        """
        if not terminal:
            return 0

        if self.cm_design:
            return self._reward_cm_design()
        elif self.rri_design:
            return self._reward_rri_design()


        # reward formulation for RNA local Design, excluding local improvement steps and gc content!!!!
        if self._env_config.local_design:
            return self.reward_local_design()
        else:
            folded_design, _ = fold(self.design.primary)
            hamming_distance = hamming(folded_design, self.target.dot_bracket)
            if 0 < hamming_distance < self._env_config.mutation_threshold:
                hamming_distance = self._local_improvement(folded_design)

            normalized_hamming_distance = hamming_distance / len(self.target)

            # For hparam optimization
            episode_info = EpisodeInfo(
                target_id=self.target.id,
                time=time.time(),
                normalized_hamming_distance=normalized_hamming_distance,
            )
            self.episodes_info.append(episode_info)

            return (1 - normalized_hamming_distance) ** self._env_config.reward_exponent

    def execute(self, actions):
        """
        Execute one interaction of the environment with the agent.

        Args:
            action: Current action of the agent.

        Returns:
            state: The next state for the agent.
            terminal: The signal for end of an episode.
            reward: The reward if at terminal timestep, else 0.
        """
        self._apply_action(actions)

        terminal = self.design.first_unassigned_site is None if not self._env_config.reward_function == 'structure_only' else all([x is None for x in self.target.structure_parts_encoding])
        state = None if terminal else self._get_state()
        reward = self._get_reward(terminal)

        return state, terminal, reward

    def close(self):
        pass

    @property
    def states(self):
        type = "int" if self._env_config.use_embedding else "float"
        if self._env_config.encoding == "multihot" and self._env_config.use_conv:
            return dict(type=type, shape=(1 + 2 * self._env_config.state_radius, 9, 1))
        if self._env_config.use_conv and not self._env_config.use_embedding:
            return dict(type=type, shape=(1 + 2 * self._env_config.state_radius, 1))
        return dict(type=type, shape=(1 + 2 * self._env_config.state_radius,))

    @property
    def actions(self):
        return dict(type="int", num_actions=4)



from typing import List, Union, Dict
import pathlib
import time, os, pickle
import numpy as np

from torchtext import data, datasets
from torchtext.data import BucketIterator, BPTTIterator

global max_src_in_batch, max_tgt_in_batch


def batch_size_fn(new, count, sofar):
    "Keep augmenting batch and calculate total number of tokens + padding."
    global max_src_in_batch, max_tgt_in_batch
    if count == 1:
        max_src_in_batch = 0
        max_tgt_in_batch = 0

    max_src_in_batch = max(max_src_in_batch, len(new.src))
    max_tgt_in_batch = max(max_tgt_in_batch, len(new.trg) + 2)
    src_elements = count * max_src_in_batch
    tgt_elements = count * max_tgt_in_batch
    return max(src_elements, tgt_elements)


class SubSortedBucketIterator(data.Iterator):
    def create_batches(self):
        if self.train:
            def pool(d, random_shuffler):
                for p in data.batch(d, self.batch_size * 100):
                    p_batch = data.batch(sorted(p, key=self.sort_key), self.batch_size, self.batch_size_fn)
                    for b in random_shuffler(list(p_batch)):
                        yield b

            self.batches = pool(self.data(), self.random_shuffler)
        else:
            self.batches = []
            for b in data.batch(self.data(), self.batch_size, self.batch_size_fn):
                self.batches.append(sorted(b, key=self.sort_key))


class BasisDataSet():

    def __init__(self, data_dir=None):

        self.blank_word = '<blank>'
        self.bos_word = '<s>'
        self.eos_word = '</s>'

        if data_dir is not None:
            self.data_dir = pathlib.Path(data_dir)
            if not os.path.isdir(".data"):
                os.mkdir(".data")

    def get_iterator(self, set_name: str, batch_size: int = None, train=False, sorted=True, bptt=False, device='cpu'):

        if batch_size is None:
            batch_size = self.batch_size
        data_set = getattr(self, set_name)

        if sorted:
            data_iter = SubSortedBucketIterator(data_set, batch_size=batch_size, device=device, repeat=False,
                                                sort_key=lambda x: (len(x.src), len(x.trg)), batch_size_fn=batch_size_fn, train=train)
        elif bptt:
            data_iter = BPTTIterator(data_set, batch_size=batch_size, device=device, repeat=False, bptt_len=bptt, train=train)
        else:
            data_iter = BucketIterator(data_set, batch_size=batch_size, device=device, repeat=False,
                                       sort_key=lambda x: (len(x.src), len(x.trg)), batch_size_fn=batch_size_fn, train=train)
        return data_iter

    def set_size(self, set_name):
        return len(getattr(self, set_name))

    @property
    def src_vocab(self):
        return self.SRC.vocab.freqs

    @property
    def trg_vocab(self):
        return self.TRG.vocab.freqs

    @property
    def src_vocab_size(self):
        return len(self.SRC.vocab)

    @property
    def trg_vocab_size(self):
        return len(self.TRG.vocab)

    def _convert_list(self, l, dictionary):
        result = []
        for t in l:
            result.append(dictionary[t])
        return result

    def _get_vocab(self, origin):
        if origin == "src":
            return self.SRC.vocab
        elif origin == "trg":
            return self.TRG.vocab
        elif origin == "dataset":
            return self.DATASET.vocab
        else:
            raise UserWarning(f"Wrong origin {origin}, expect src or trg")

    def _get_stoi(self, origin):
        if origin == "src":
            return self.SRC.vocab.stoi
        elif origin == "trg":
            return self.TRG.vocab.stoi
        elif origin == "dataset":
            return self.DATASET.vocab.stoi
        else:
            raise UserWarning(f"Wrong origin {origin}, expect src or trg")

    def stoi(self, s: Union[List[str], str], origin: str) -> Union[List[int], int]:
        vocab = self._get_vocab(origin)
        if isinstance(s, str):
            return vocab.stoi[s]
        elif isinstance(s, List):
            return self._convert_list(s, vocab.stoi)
        else:
            raise UserWarning(f"Wrong data type: {type(s)}, expect int or List")

    def itos(self, i: Union[List[int], int], origin: str) -> Union[List[str], str]:
        vocab = self._get_vocab(origin)
        if isinstance(i, int):
            return vocab.itos[i]
        elif isinstance(i, List):
            return self._convert_list(i, vocab.itos)
        else:
            raise UserWarning(f"Wrong data type: {type(i)}, expect int or List")

    def add_bos_eos(self, l: List[Union[int, str]], origin: str) -> List[Union[int, str]]:
        if isinstance(l[0], int):
            bos_word = self.stoi(self.bos_word, origin)
            eos_word = self.stoi(self.eos_word, origin)
            blank_word = self.stoi(self.blank_word, origin)
        elif isinstance(l[0], str):
            bos_word = self.bos_word
            eos_word = self.eos_word
            blank_word = self.blank_word
        else:
            raise UserWarning(f"Wrong data type in list: {type(l[0])}, expect list elements are int or str")

        if l[0] != bos_word:
            l = [bos_word] + l
        if not eos_word in l:
            if blank_word in l:
                blank_idx = l.index(blank_word)
                l[blank_idx] = eos_word
            else:
                l = l + [eos_word, ]
        return l

    def remove_bos_eos(self, l: List[Union[int, str]], origin: str) -> List[Union[int, str]]:
        if isinstance(l[0], int):
            bos_word = self.stoi(self.bos_word, origin)
            eos_word = self.stoi(self.eos_word, origin)
            blank_word = self.stoi(self.blank_word, origin)
        elif isinstance(l[0], str):
            bos_word = self.bos_word
            eos_word = self.eos_word
            blank_word = self.blank_word
        else:
            raise UserWarning(f"Wrong data type in list: {type(l[0])}, expect list elements are int or str")

        if l[0] == bos_word:
            l = l[1:]
        if eos_word in l:
            eos_idx = l.index(eos_word)
            l = l[:eos_idx]
        if blank_word in l:
            blank_idx = l.index(blank_word)
            l = l[:blank_idx]
        return l



from tensorforce.agents import PPOAgent, RandomAgent, TRPOAgent
from dataclasses import dataclass

from functools import partial


@dataclass
class NetworkConfig:
    """
    Dataclass providing the network configuration.

    Default values describe:
        conv_sizes: The number of filters.
        conv_channels: The convolution window size.
        lstm_units: The number of lstm units in a single layer.
        fc_units: The number of units in a single fully-connected layer.
        fc_activation: The activation function.
        num_fc_layers: The number of fully-connected layers.
        embedding_size: The dimensionality of the embedding layer.
    """

    conv_sizes: int = (3, 2)
    conv_channels: int = (16, 8)
    lstm_units: int = 1
    num_lstm_layers: int = 0
    fc_units: int = 2
    fc_activation: str = "relu"
    num_fc_layers: int = 2
    embedding_size: int = 0
    conv2d: bool = False
    embedding_indices: int = 9


def get_network(network_config):
    """
    Get a specific policy network as specified in the network configuration.

    Args:
        network_config: The configuration for the nertwork.

    Returns:
        The policy network of the agent.
    """
    embedding = [
        dict(
            type="embedding",
            indices=network_config.embedding_indices,
            size=network_config.embedding_size,
            l2_regularization=0.0,
            l1_regularization=0.0,
        )
    ]
    if not network_config.conv2d:
        convolution = [
            dict(
                type="conv1d",
                size=size,
                window=window,
                stride=1,
                padding="VALID",
                bias=True,
                activation="relu",
                l2_regularization=0.0,
                l1_regularization=0.0,
            )
            for size, window in zip(network_config.conv_channels, network_config.conv_sizes)
            if window > 1
        ]
    else:
        convolution = [
        dict(
            type="conv2d",
            size=size,
            window=window,
            stride=1,
            padding="SAME",
            bias=True,
            activation="relu",
            l2_regularization=0.0,
            l1_regularization=0.0,
        )
        for size, window in zip(network_config.conv_channels, network_config.conv_sizes)
        if window > 1
        ]
    lstm = [dict(type="internal_lstm", size=network_config.lstm_units)]
    dense = [
        dict(
            type="dense",
            size=network_config.fc_units,
            bias=True,
            activation=network_config.fc_activation,
            l2_regularization=0.0,
            l1_regularization=0.0,
        )
    ]

    use_conv = any(map(lambda x: x > 1, network_config.conv_sizes))
    network = []
    if network_config.embedding_size:
        network += embedding
    if use_conv:
        network += convolution
    if use_conv or network_config.embedding_size:
        network += [dict(type="flatten")]
    network += network_config.num_lstm_layers * lstm
    network += network_config.num_fc_layers * dense
    return network


@dataclass
class AgentConfig:
    """
    Dataclass providing the agent configuration.

    Default values describe:
        learning_rate: The learning rate to use by PPO.
        batch_size: Integer of the batch size.
        optimization_steps: The number of optimization steps.
        likelihood_ratio_clipping: Likelihood ratio clipping for policy gradient.
        entropy_regularization: Entropy regularization weight.
        random_agent: Defines if agent is a random agent.
    """

    learning_rate: float = 5e-4
    batch_size: int = 5
    likelihood_ratio_clipping: float = 0.3
    entropy_regularization: float = 1.5e-3
    agent: str = 'ppo'


# This is needed / used because of the threaded runner interface of tensorforce, see
# learn_to_design_rna.py.
def ppo_agent_kwargs(agent_config, session_config):
    """
    Get keyword arguments for initializing a PPO agent.

    Args:
        agent_config: The configuration of the agent.
        session_config: The session configuration.

    Returns:
        Dictionary of arguments for initialization of a PPO agent.
    """
    step_optimizer = dict(type="adam", learning_rate=agent_config.learning_rate)
    return dict(
        device=None,
        session_config=session_config,
        distributed_spec=None,
        distributions_spec=None,
        discount=1.0,
        batched_observe=1000,  # set to batch_size?
        batch_size=agent_config.batch_size,
        keep_last_timestep=True,
        step_optimizer=step_optimizer,
        optimization_steps=1,
        likelihood_ratio_clipping=agent_config.likelihood_ratio_clipping,
        entropy_regularization=agent_config.entropy_regularization,
        variable_noise=None,
        states_preprocessing_spec=None,
        explorations_spec=None,
        reward_preprocessing_spec=None,
        baseline_mode=None,
        baseline=None,
        baseline_optimizer=None,
        gae_lambda=None,
    )

def trpo_agent_kwargs(agent_config, session_config):
    """
    Get keyword arguments for initializing a PPO agent.

    Args:
        agent_config: The configuration of the agent.
        session_config: The session configuration.

    Returns:
        Dictionary of arguments for initialization of a TRPO agent.
    """
    # baseline_optimizer = dict(type="adam", learning_rate=agent_config.learning_rate)
    return dict(
        device=None,
        session_config=None,
        scope='trpo',
        saver_spec=None,
        summary_spec=None,
        distributed_spec=None,
        discount=0.99,
        variable_noise=None,
        states_preprocessing_spec=None,
        explorations_spec=None,
        reward_preprocessing_spec=None,
        distributions_spec=None,
        entropy_regularization=None,
        baseline_mode=None,
        baseline=None,
        baseline_optimizer=None,
        gae_lambda=None,
        batched_observe=1000,
        batch_size=1000,
        keep_last_timestep=True,
        likelihood_ratio_clipping=None,
        learning_rate=1e-3,
        cg_max_iterations=20,
        cg_damping=1e-3,
        cg_unroll_loop=False
        )

# This is needed / used because of the threaded runner interface of tensorforce, see
# learn_to_design_rna.py.
def random_agent_kwargs(agent_config, session_config):
    """
    Get a random agent's keyword arguments for initialization.

    Args:
        agent_config: The configuration of the agent.
        session_config: The session configuration.

    Returns:
        A dictionary of keyword args for a random agent.
    """
    return dict(
        device=None,
        session_config=session_config,
        distributed_spec=None,
        discount=1.0,
        batched_observe=1000,
        variable_noise=None,
        states_preprocessing_spec=None,
        explorations_spec=None,
        reward_preprocessing_spec=None,
    )


def get_agent(environment, network, agent_config, session_config, restore_path):
    """
    Get an agent. General function providing agents.

    Args:
        environment: The environment.
        network: The network of the agent.
        agent_config: The configuration of the agent.
        session_config: The configuration of the session.
        restore_path: Path to restore saved data from.

    Returns:
       An agent.
    """
    if agent_config.agent == 'random':
        return RandomAgent(
            environment.states,
            environment.actions,
            **random_agent_kwargs(agent_config, session_config)
        )
    elif agent_config.agent == 'trpo':
        agent = TRPOAgent(
            environment.states,
            environment.actions,
            network,
            **trpo_agent_kwargs(agent_config, session_config)
        )
    else:
        agent = PPOAgent(
            environment.states,
            environment.actions,
            network,
            **ppo_agent_kwargs(agent_config, session_config)
        )
    if restore_path:
        agent.restore_model(directory=restore_path)
    return agent


def get_agent_fn(**kwargs):
    """
    Function to generically get agents using keyword arguments <**kwargs>.
    """
    return partial(get_agent, **kwargs)
